# (c) Kevin Dunn, 2010-2021. MIT License. Based on own private work over the years.
import time
from functools import partial
from typing import Optional, Any
import warnings

import numpy as np
import pandas as pd
from scipy.stats import f, chi2

from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.decomposition import PCA as PCA_sklearn
from sklearn.cross_decomposition import PLSRegression as PLS_sklearn
from sklearn.utils.validation import check_is_fitted

epsqrt = np.sqrt(np.finfo(float).eps)


class SpecificationWarning(UserWarning):
    """Parent warning class."""

    pass


class MCUVScaler(BaseEstimator, TransformerMixin):
    """
    Create our own mean centering and scaling to unit variance (MCUV) class
    The default scaler in sklearn does not handle small datasets accurately, with ddof.
    """

    def __init__(self):
        pass

    def fit(self, X, y=None):
        self.center_ = X.mean()
        # this is the key difference with "preprocessing.StandardScaler"
        self.scale_ = X.std(ddof=1)
        self.scale_[self.scale_ == 0] = 1.0  # columns with no variance are left as-is.
        return self

    def transform(self, X):
        check_is_fitted(self, "center_")
        check_is_fitted(self, "scale_")

        X = X.copy()
        return (X - self.center_) / self.scale_

    def inverse_transform(self, X):
        check_is_fitted(self, "center_")
        check_is_fitted(self, "scale_")

        X = X.copy()
        return X * self.scale_ + self.center_


class PCA(PCA_sklearn):
    def __init__(
        self,
        n_components=None,
        *,
        copy=True,
        whiten=False,
        svd_solver="auto",
        tol=0.0,
        iterated_power="auto",
        random_state=None,
        # Own extra inputs, for the case when there is missing data
        missing_data_settings: Optional[dict] = None,
    ):
        super().__init__(
            n_components=n_components,
            copy=copy,
            whiten=whiten,
            svd_solver=svd_solver,
            tol=tol,
            iterated_power=iterated_power,
            random_state=random_state,
        )
        self.n_components: int = n_components
        self.missing_data_settings = missing_data_settings
        self.has_missing_data = False

    def fit(self, X, y=None) -> PCA_sklearn:
        """
        Fits a principal component analysis (PCA) model to the data.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training data, where `n_samples` is the number of samples (rows)
            and `n_features` is the number of features (columns).
        y : Not used; by default None

        Returns
        -------
        PCA
            Model object.
        """
        self.N, self.K = X.shape

        # Check if number of components is supported against maximum requested
        min_dim = int(min(self.N, self.K))
        self.A: int = min_dim if self.n_components is None else int(self.n_components)

        if self.A > min_dim:
            warn = (
                "The requested number of components is more than can be "
                "computed from data. The maximum number of components is "
                f"the minimum of either the number of rows ({self.N}) or "
                f"the number of columns ({self.K})."
            )
            warnings.warn(warn, SpecificationWarning)
            self.A = self.n_components = min_dim

        if np.any(X.isna()):
            # If there are missing data, then the missing data settings apply. Defaults are:
            #
            #       md_method = "pmp"
            #       md_tol = np.sqrt(np.finfo(float).eps)
            #       md_max_iter = 1000

            default_mds = dict(md_method="tsr", md_tol=epsqrt, md_max_iter=1000)
            if isinstance(self.missing_data_settings, dict):
                default_mds.update(self.missing_data_settings)

            self.missing_data_settings = default_mds
            self = PCA_missing_values(
                n_components=self.n_components,
                missing_data_settings=self.missing_data_settings,
            )
            self.N, self.K = X.shape
            self.A = self.n_components
            self.fit(X)

        else:
            self = super().fit(X)
            self.x_loadings = self.components_.T
            self.extra_info = {}
            self.extra_info["timing"] = np.zeros((1, self.A)) * np.nan
            self.extra_info["iterations"] = np.zeros((1, self.A)) * np.nan

        # We have now fitted the model. Apply some convenience shortcuts for the user.
        self.A = self.n_components
        self.N = self.n_samples_
        self.K = self.n_features_

        self.loadings = pd.DataFrame(self.x_loadings)
        self.loadings.index = X.columns

        component_names = [f"{a+1}" for a in range(self.A)]
        self.loadings.columns = component_names

        self.scaling_factor_for_scores = pd.Series(
            np.sqrt(self.explained_variance_),
            index=component_names,
            name="Standard deviation per score",
        )
        self.Hotellings_T2 = pd.DataFrame(
            np.zeros(shape=(self.N, self.A)),
            columns=component_names,
            index=X.index,
            # name="Hotelling's T^2 statistic, per component",
        )
        if self.has_missing_data:
            self.x_scores = pd.DataFrame(self.x_scores_, columns=component_names, index=X.index)
            self.squared_prediction_error = pd.DataFrame(
                self.squared_prediction_error_,
                columns=component_names,
                index=X.index.copy(),
            )
            self.R2 = pd.Series(
                self.R2_,
                index=component_names,
                name="Model's R^2, per component",
            )
            self.R2cum = pd.Series(
                self.R2cum_,
                index=component_names,
                name="Cumulative model's R^2, per component",
            )
            self.R2X_k_cum = pd.DataFrame(
                self.R2X_k_cum_,
                columns=component_names,
                index=X.columns,
                # name ="Per variable R^2, per component"
            )
        else:
            self.x_scores = pd.DataFrame(
                super().fit_transform(X), columns=component_names, index=X.index
            )
            self.squared_prediction_error = pd.DataFrame(
                np.zeros((self.N, self.A)),
                columns=component_names,
                index=X.index.copy(),
            )
            self.R2 = pd.Series(
                np.zeros(shape=(self.A,)),
                index=component_names,
                name="Model's R^2, per component",
            )
            self.R2cum = pd.Series(
                np.zeros(shape=(self.A,)),
                index=component_names,
                name="Cumulative model's R^2, per component",
            )
            self.R2X_k_cum = pd.DataFrame(
                np.zeros(shape=(self.K, self.A)),
                columns=component_names,
                index=X.columns,
                # name ="Per variable R^2, per component"
            )

        if not self.has_missing_data:
            Xd = X.copy()
            prior_SSX_col = ssq(Xd.values, axis=0)
            base_variance = np.sum(prior_SSX_col)
            for a in range(self.A):
                Xd -= self.x_scores.iloc[:, [a]] @ self.loadings.iloc[:, [a]].T
                # These are the Residual Sums of Squares (RSS); i.e X-X_hat
                row_SSX = ssq(Xd.values, axis=1)
                col_SSX = ssq(Xd.values, axis=0)

                # Don't use a check correction factor. Define SPE simply as the sum of squares of
                # the errors, then take the square root, so it is interpreted like a standard error.
                # If the user wants to normalize it, then this is a clean base value to start from.
                self.squared_prediction_error.iloc[:, a] = np.sqrt(row_SSX)

                # TODO: some entries in prior_SSX_col can be zero and leads to nan's in R2X_k_cum
                self.R2X_k_cum.iloc[:, a] = 1 - col_SSX / prior_SSX_col

                # R2 and cumulative R2 value for the whole block
                self.R2cum[a] = 1 - sum(row_SSX) / base_variance
                if a > 0:
                    self.R2[a] = self.R2cum[a] - self.R2cum[a - 1]
                else:
                    self.R2[a] = self.R2cum[a]
        # end: has no missing data

        for a in range(self.A):
            self.Hotellings_T2.iloc[:, a] = (
                self.Hotellings_T2.iloc[:, max(0, a - 1)]
                + (self.x_scores.iloc[:, a] / self.scaling_factor_for_scores[a]) ** 2
            )

        self.ellipse_coordinates = partial(
            ellipse_coordinates,
            n_components=self.n_components,
            scaling_factor_for_scores=self.scaling_factor_for_scores,
            n_rows=self.N,
        )

        self.T2_limit = partial(T2_limit, n_components=self.n_components, n_rows=self.N)

        return self

    def fit_transform(self, X, y=None):
        self.fit(X)
        assert False, "Still do the transform part"

    def SPE_limit(self, conf_level=0.95) -> float:
        check_is_fitted(self, "squared_prediction_error")

        assert conf_level > 0.0
        assert conf_level < 1.0

        # The limit is for the squares (i.e. the sum of the squared errors)
        # I.e. self.squared_prediction_error was square-rooted outside this function, so undo that.
        values = self.squared_prediction_error.iloc[:, self.A - 1] ** 2

        center_spe = values.mean()
        variance_spe = values.var(ddof=1)

        g = variance_spe / (2 * center_spe)
        h = (2 * (center_spe ** 2)) / variance_spe
        # Then take the square root again, to return the limit for SPE
        return np.sqrt(chi2.ppf(conf_level, h) * g)


class PCA_missing_values(BaseEstimator, TransformerMixin):
    """
    Create our PCA class if there is even a single missing data value in the X input array.

    The default method to impute missing values is the TSR algorithm (`md_method="tsr"`).

    Missing data method options are:

    * 'pmp'         Projection to Model Plane
    * 'scp'         Single Component Projection
    * 'nipals'      Same as 'scp': non-linear iterative partial least squares.
    * 'tsr': TODO
    * Other options? See papers by Abel Folch-Fortuny and also:
    https://analyticalsciencejournals.onlinelibrary.wiley.com/doi/abs/10.1002/cem.750

    """

    valid_md_methods = ["pmp", "scp", "nipals", "tsr"]

    def __init__(
        self,
        n_components=None,
        copy: bool = True,
        missing_data_settings=dict,
    ):
        self.n_components = n_components
        self.missing_data_settings = missing_data_settings
        self.has_missing_data = True
        self.missing_data_settings["md_max_iter"] = int(self.missing_data_settings["md_max_iter"])

        assert self.missing_data_settings["md_tol"] < 10, "Tolerance should not be too large"
        assert (
            self.missing_data_settings["md_tol"] > epsqrt ** 1.95
        ), "Tolerance must exceed machine precision"

        assert self.missing_data_settings["md_method"] in self.valid_md_methods, (
            f"Missing data method is not recognized. Must be one of {self.valid_md_methods}.",
        )

    def fit(self, X, y=None):

        # Force input to NumPy array:
        self.data = np.asarray(X)

        # Other setups:
        self.n_components = self.A
        self.n_samples_ = self.N
        self.n_features_ = self.K

        self.x_loadings = np.zeros((self.K, self.A))
        self.x_scores_ = np.zeros((self.N, self.A))
        self.R2_ = np.zeros(shape=(self.A,))
        self.R2cum_ = np.zeros(shape=(self.A,))
        self.R2X_k_cum_ = np.zeros(shape=(self.K, self.A))
        self.squared_prediction_error_ = np.zeros((self.N, self.A))

        # Perform MD algorithm here
        if self.missing_data_settings["md_method"].lower() == "pmp":
            self._fit_pmp(X)
        elif self.missing_data_settings["md_method"].lower() in ["scp", "nipals"]:
            self._fit_nipals_pca(settings=self.missing_data_settings)
        elif self.missing_data_settings["md_method"].lower() in ["tsr"]:
            self._fit_tsr(settings=self.missing_data_settings)

        # Additional calculations, which can be done after the missing data method is complete.
        self.explained_variance_ = np.diag(self.x_scores_.T @ self.x_scores_) / (self.N - 1)
        return self

    def transform(self, X):
        check_is_fitted(self, "blah")

        X = X.copy()
        return X

    def inverse_transform(self, X):
        check_is_fitted(self, "blah")

        X = X.copy()
        return X

    def _fit_nipals_pca(self, settings):
        """
        Internal method to fit the PCA model using the NIPALS algorithm.
        """
        # NIPALS algorithm
        K, A = self.K, self.A

        # Create direct links to the data
        Xd = np.asarray(self.data)
        base_variance = ssq(Xd)

        # Initialize storage:
        self.extra_info = {}
        self.extra_info["timing"] = np.zeros(A) * np.nan
        self.extra_info["iterations"] = np.zeros(A) * np.nan

        for a in np.arange(A):

            # Timers and housekeeping
            start_time = time.time()
            itern = 0
            start_SS_col = ssq(Xd, axis=0)

            if sum(start_SS_col) < epsqrt:
                emsg = (
                    "There is no variance left in the data array: cannot "
                    f"compute any more components beyond component {a}."
                )
                raise RuntimeError(emsg)

            # Initialize t_a with random numbers, or carefully select a column
            # from X. <-- Don't do this anymore.
            # Rather: Pick a column from X as the initial guess instead.
            t_a_guess = Xd[:, [0]]  # .reshape(self.N, 1)
            t_a_guess[np.isnan(t_a_guess)] = 0
            t_a = t_a_guess + 1.0
            p_a = np.zeros((K, 1))
            while not (terminate_check(t_a_guess, t_a, iterations=itern, settings=settings)):

                # 0: Richardson's acceleration, or any numerical acceleration
                #    method for PCA where there is slow convergence?

                # 0: starting point for convergence checking on next loop
                t_a_guess = t_a.copy()

                # 1: Regress the score, t_a, onto every column in X, compute the
                #    regression coefficient and store in p_a
                # p_a = X.T * t_a / (t_a.T * t_a)
                # p_a = (X.T)(t_a) / ((t_a.T)(t_a))
                # p_a = np.dot(X.T, t_a) / ssq(t_a)
                p_a = quick_regress(Xd, t_a)

                # 2: Normalize p_a to unit length
                p_a /= np.sqrt(ssq(p_a))

                # 3: Now regress each row in X on the p_a vector, and store the
                #    regression coefficient in t_a
                # t_a = X * p_a / (p_a.T * p_a)
                # t_a = (X)(p_a) / ((p_a.T)(p_a))
                # t_a = np.dot(X, p_a) / ssq(p_a)
                t_a = quick_regress(Xd, p_a)

                itern += 1

            self.extra_info["timing"][a] = time.time() - start_time
            self.extra_info["iterations"][a] = itern

            # Loop terminated!  Now deflate the X-matrix
            Xd -= np.dot(t_a, p_a.T)
            # These are the Residual Sums of Squares (RSS); i.e X-X_hat
            row_SSX = ssq(Xd, axis=1)
            col_SSX = ssq(Xd, axis=0)

            self.squared_prediction_error_[:, a] = np.sqrt(row_SSX)

            # TODO: some entries in start_SS_col can be zero and leads to nan's in R2X_k_cum
            self.R2X_k_cum_[:, a] = 1 - col_SSX / start_SS_col

            # R2 and cumulative R2 value for the whole block
            self.R2cum_[a] = 1 - sum(row_SSX) / base_variance
            if a > 0:
                self.R2_[a] = self.R2cum_[a] - self.R2cum_[a - 1]
            else:
                self.R2_[a] = self.R2cum_[a]

            # VIP value (only calculated for X-blocks); only last column is useful
            # self.VIP_a = np.zeros((self.K, self.A))
            # self.VIP = np.zeros(self.K)

            # Store results
            # -------------
            # Flip the signs of the column vectors in P so that the largest
            # magnitude element is positive
            # (Wold, Esbensen, Geladi, PCA,  CILS, 1987, p 42)
            # http://dx.doi.org/10.1016/0169-7439(87)80084-9
            max_el_idx = np.argmax(np.abs(p_a))
            if np.sign(p_a[max_el_idx]) < 1:
                p_a *= -1.0
                t_a *= -1.0

            # Store the loadings and scores
            self.x_loadings[:, a] = p_a.flatten()
            self.x_scores_[:, a] = t_a.flatten()

        # end looping on A components

    def _fit_tsr(self, settings):
        delta = 1e100
        Xd = np.asarray(self.data)
        N, K, A = self.N, self.K, self.A
        mmap = np.isnan(Xd)
        # Could impute missing values per row, but leave it at zero, assuming the data are MCUV.
        Xd[mmap] = 0.0
        itern = 0
        while (itern < settings["md_max_iter"]) and (delta > settings["md_tol"]):
            itern += 1
            missing_X = Xd[mmap]
            mean_X = np.mean(Xd, axis=0)
            S = np.cov(Xd, rowvar=False, ddof=1)
            Xc = Xd - mean_X
            if N > K:
                _, _, V = np.linalg.svd(Xc, full_matrices=False)
            else:
                V, _, _ = np.linalg.svd(Xc.T, full_matrices=False)

            V = V.T[:, 0:A]  # transpose first
            for n in range(N):
                # If there are missing values (mis) in the n-th row. Compared to obs-erved values.
                row_mis = mmap[n, :]
                row_obs = ~row_mis
                if np.any(row_mis):
                    # Form the so-called key-matrix, L
                    L = V[row_obs, 0 : min(A, sum(row_obs))]
                    S11 = S[row_obs, :][:, row_obs]
                    S21 = S[row_mis, :][:, row_obs]
                    z2 = (S21 @ L) @ np.linalg.pinv(L.T @ S11 @ L) @ L.T
                    Xc[n, row_mis] = z2 @ Xc[n, row_obs]
            Xd = Xc + mean_X
            delta = np.mean((Xd[mmap] - missing_X) ** 2)

        # All done: return the results in `self`
        S = np.cov(Xd, rowvar=False, ddof=1)
        _, _, V = np.linalg.svd(S, full_matrices=False)
        self.x_loadings = (V[0:A, :]).T  # transpose result to the right shape: K x A
        self.x_scores_ = (Xd - np.mean(Xd, axis=0)) @ self.x_loadings
        self.data = Xd


class PLS(PLS_sklearn):
    def __init__(
        self,
        n_components: int = 2,
        *,
        scale: bool = True,
        max_iter: int = 1000,
        tol: float = epsqrt,
        copy: bool = True,
        # Own extra inputs, for the case when there is missing data
        missing_data_settings: Optional[dict] = None,
    ):
        super().__init__(
            n_components=n_components,
            scale=scale,
            max_iter=max_iter,
            tol=tol,
            copy=copy,
        )
        self.n_components: int = n_components
        self.missing_data_settings = missing_data_settings
        self.has_missing_data = False

    def fit(self, X, Y) -> PLS_sklearn:
        """
        Fits a projection to latent structures (PLS) or Partial Least Square (PLS) model to the
        data.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training data, where `n_samples` is the number of samples (rows)
            and `n_features` is the number of features (columns).
        Y : array-like, shape (n_samples, n_targets)
            Training data, where `n_samples` is the number of samples (rows)
            and `n_targets` is the number of target outputs (columns).

        Returns
        -------
        PLS
            Model object.

        References
        ----------

        Abdi, "Partial least squares regression and projection on latent structure
        regression (PLS Regression)", 2010, DOI: 10.1002/wics.51
        """
        self.N, self.K = X.shape
        self.Ny, self.M = Y.shape
        assert self.Ny == self.N, (
            f"The X and Y arrays must have the same number of rows: X has {self.N} and "
            f"Y has {self.Ny}."
        )

        # Check if number of components is supported against maximum requested
        min_dim = min(self.N, self.K)
        self.A = min_dim if self.n_components is None else int(self.n_components)
        if self.A > min_dim:
            warn = (
                "The requested number of components is more than can be "
                "computed from data. The maximum number of components is "
                f"the minimum of either the number of rows ({self.N}) or "
                f"the number of columns ({self.K})."
            )
            warnings.warn(warn, SpecificationWarning)
            self.A = self.n_components = min_dim

        if np.any(Y.isna()) or np.any(X.isna()):
            # If there are missing data, then the missing data settings apply. Defaults are:
            #
            #       md_method = "tsr"
            #       md_tol = np.sqrt(np.finfo(float).eps)
            #       md_max_iter = self.max_iter

            default_mds = dict(md_method="tsr", md_tol=epsqrt, md_max_iter=self.max_iter)
            if isinstance(self.missing_data_settings, dict):
                default_mds.update(self.missing_data_settings)

            self.missing_data_settings = default_mds
            self = PLS_missing_values(
                n_components=self.n_components,
                missing_data_settings=self.missing_data_settings,
            )
            self.N, self.K = X.shape
            self.Ny, self.M = Y.shape
            self.A = self.n_components

            # Call the sub-function to do the PLS fit when missing data are present
            self.fit(X, Y)
        else:
            self = super().fit(X, Y)
            # x_scores #T: N x A
            # y_scores # U: N x A
            # x_weights # W: K x A
            # y_weights # identical values to y_loadings. So this is redundant then.
            # x_loadings  # P: K x A
            # y_loadings  # C: M x A

            self.extra_info = {}
            self.extra_info["timing"] = np.zeros(self.A) * np.nan
            self.extra_info["iterations"] = np.array(self.n_iter_)

        # We have now fitted the model. Apply some convenience shortcuts for the user.
        self.A = self.n_components

        # Further convenience calculations
        # "direct_weights" is matrix R = W(P'W)^{-1}  [R is KxA matrix]; useful since T = XR
        direct_weights = self.x_weights_ @ np.linalg.inv(self.x_loadings_.T @ self.x_weights_)
        # beta = RC' [KxA by AxM = KxM]: the KxM matrix gives the direct link from the k-th (input)
        # X variable, to the m-th (output) Y variable.
        beta_coefficients = direct_weights @ self.y_loadings_.T

        # Initialize storage:
        component_names = [f"{a+1}" for a in range(self.A)]
        self.x_scores = pd.DataFrame(self.x_scores_, index=X.index, columns=component_names)
        self.y_scores = pd.DataFrame(self.y_scores_, index=Y.index, columns=component_names)
        self.x_weights = pd.DataFrame(self.x_weights_, index=X.columns, columns=component_names)
        self.y_weights = pd.DataFrame(self.y_weights_, index=Y.columns, columns=component_names)
        self.x_loadings = pd.DataFrame(self.x_loadings_, index=X.columns, columns=component_names)
        self.y_loadings = pd.DataFrame(self.y_loadings_, index=Y.columns, columns=component_names)
        self.predictions = pd.DataFrame(
            self.x_scores @ self.y_loadings.T, index=Y.index, columns=Y.columns
        )
        self.direct_weights = pd.DataFrame(direct_weights, index=X.columns, columns=component_names)
        self.beta_coefficients = pd.DataFrame(beta_coefficients, index=X.columns, columns=Y.columns)
        self.explained_variance = np.diag(self.x_scores.T @ self.x_scores) / (self.N - 1)
        self.scaling_factor_for_scores = pd.Series(
            np.sqrt(self.explained_variance),
            index=component_names,
            name="Standard deviation per score",
        )
        self.Hotellings_T2 = pd.DataFrame(
            np.zeros(shape=(self.N, self.A)),
            columns=component_names,
            index=X.index,
            # name="Hotelling's T^2 statistic, per component",
        )
        self.squared_prediction_error = pd.DataFrame(
            np.zeros((self.N, self.A)),
            columns=component_names,
            index=X.index.copy(),
        )
        self.R2 = pd.Series(
            np.zeros(shape=(self.A)),
            index=component_names,
            name="Model's R^2, per component",
        )
        self.R2cum = pd.Series(
            np.zeros(shape=(self.A)),
            index=component_names,
            name="Cumulative model's R^2, per component",
        )
        self.R2X_k_cum = pd.DataFrame(
            np.zeros(shape=(self.K, self.A)),
            index=X.columns,
            columns=component_names,
            # name ="Per variable in the X-space: R^2, per component"
        )
        self.R2Y_m_cum = pd.DataFrame(
            np.zeros(shape=(self.M, self.A)),
            index=Y.columns,
            columns=component_names,
            # name ="Per variable in the Y-space: R^2, per component"
        )

        Xd = X.copy()
        Yd = Y.copy()
        prior_SSX_col = ssq(Xd.values, axis=0)
        prior_SSY_col = ssq(Yd.values, axis=0)
        base_variance_Y = np.sum(prior_SSY_col)
        for a in range(self.A):
            self.Hotellings_T2[f"{a+1}"] = (
                self.Hotellings_T2.iloc[:, max(0, a - 1)]
                + (self.x_scores.iloc[:, a] / self.scaling_factor_for_scores[a]) ** 2
            )
            Xd -= self.x_scores.iloc[:, [a]] @ self.x_loadings.iloc[:, [a]].T
            y_hat = self.x_scores.iloc[:, 0 : (a + 1)] @ self.y_loadings.iloc[:, 0 : (a + 1)].T
            # These are the Residual Sums of Squares (RSS); i.e X-X_hat
            row_SSX = ssq(Xd.values, axis=1)
            col_SSX = ssq(Xd.values, axis=0)
            row_SSY = ssq(y_hat.values, axis=1)
            col_SSY = ssq(y_hat.values, axis=0)
            # R2 and cumulative R2 value for the whole block
            self.R2cum[a] = sum(row_SSY) / base_variance_Y
            if a > 0:
                self.R2[a] = self.R2cum[a] - self.R2cum[a - 1]
            else:
                self.R2[a] = self.R2cum[a]

            # Don't use a check correction factor. Define SPE simply as the sum of squares of
            # the errors, then take the square root, so it is interpreted like a standard error.
            # If the user wants to normalize it, then this is a clean base value to start from.
            self.squared_prediction_error.iloc[:, a] = np.sqrt(row_SSX)

            # TODO: some entries in prior_SSX_col can be zero and leads to nan's in R2X_k_cum
            self.R2X_k_cum.iloc[:, a] = 1 - col_SSX / prior_SSX_col
            self.R2Y_m_cum.iloc[:, a] = 1 - col_SSY / prior_SSY_col

        self.y_predicted = y_hat

        self.ellipse_coordinates = partial(
            ellipse_coordinates,
            n_components=self.n_components,
            scaling_factor_for_scores=self.scaling_factor_for_scores,
            n_rows=self.N,
        )

        self.T2_limit = partial(T2_limit, n_components=self.n_components, n_rows=self.N)

        return self

    def predict(self, X):
        """
        Using the PLS model on new data coming in matrix X.
        """

        class State(object):
            """Class object to hold the prediction results together."""

            pass

        state = State()
        state.N, state.K = X.shape
        assert self.K == state.K, "Prediction data must same number of columns as training data."

        state.x_scores = X @ self.direct_weights

        # TODO: handle the missing data version here still
        for a in range(self.A):
            pass
            # p = self.x_loadings.iloc[:, [a]]
            # w = self.x_weights.iloc[:, [a]]
            # temp = X @ self.direct_weights.iloc[:, [a]]
            # X_mcuv -= temp @ p.T
            # state.x_scores[:, [a]] = temp

        # Scores are calculated, now do the rest
        state.Hotellings_T2 = np.sum(
            np.power((state.x_scores / self.scaling_factor_for_scores.values), 2), 1
        )
        # Calculate SPE-residuals (sum over rows of the errors)
        X_mcuv = X.copy()
        X_mcuv -= state.x_scores @ self.x_loadings.T
        state.squared_prediction_error = np.sqrt(np.power(X_mcuv, 2).sum(axis=1))
        # Predicted values from the model (user still has to un-preprocess these predictions!)
        state.y_hat = state.x_scores @ self.y_loadings.T

        return state


class PLS_missing_values(BaseEstimator, TransformerMixin):
    """
    Create our PLS class if there is even a single missing data value in the X input array.

    The default method to impute missing values is the TSR algorithm (`md_method="tsr"`).

    Missing data method options are:

    * 'pmp'         Projection to Model Plane
    * 'scp'         Single Component Projection
    * 'nipals'      Same as 'scp': non-linear iterative partial least squares.
    * 'tsr':        Trimmed score regression

    * Other options? See papers by Abel Folch-Fortuny and also:
    https://analyticalsciencejournals.onlinelibrary.wiley.com/doi/abs/10.1002/cem.750

    """

    valid_md_methods = ["pmp", "scp", "nipals", "tsr"]

    def __init__(
        self,
        n_components=None,
        copy: bool = True,
        missing_data_settings=dict,
    ):
        self.n_components = n_components
        self.missing_data_settings = missing_data_settings
        self.has_missing_data = True
        self.missing_data_settings["md_max_iter"] = int(self.missing_data_settings["md_max_iter"])

        assert self.missing_data_settings["md_tol"] < 10, "Tolerance should not be too large"
        assert (
            self.missing_data_settings["md_tol"] > epsqrt ** 1.95
        ), "Tolerance must exceed machine precision"

        assert self.missing_data_settings["md_method"] in self.valid_md_methods, (
            f"Missing data method is not recognized. Must be one of {self.valid_md_methods}.",
        )

    def fit(self, X, Y):
        """
        Fits a PLS latent variable model between `X` and `Y` data arrays, accounting for missing
        values (nan's) in either or both arrays.

        1.  Höskuldsson, PLS regression methods, Journal of Chemometrics, 2(3), 211-228, 1998,
            http://dx.doi.org/10.1002/cem.1180020306
        """
        # Force input to NumPy array:
        self.Xd = np.asarray(X)
        self.Yd = np.asarray(Y)

        if np.any(np.sum(self.Yd, axis=1) == 0):
            raise Warning(
                (
                    "Cannot handle the case yet where the entire observation in Y-matrix is "
                    "missing. Please remove those rows and refit model."
                )
            )

        # Other setups:
        self.n_components = self.A
        self.n_samples_ = self.N
        self.n_features_ = self.K
        # self.M ?

        self.x_scores_ = np.zeros((self.N, self.A))  # T: N x A
        self.y_scores_ = np.zeros((self.N, self.A))  # U: N x A
        self.x_weights_ = np.zeros((self.K, self.A))  # W: K x A
        self.y_weights_ = None
        self.x_loadings_ = np.zeros((self.K, self.A))  # P: K x A
        self.y_loadings_ = np.zeros((self.M, self.A))  # C: M x A

        self.R2 = np.zeros(shape=(self.A,))
        self.R2cum = np.zeros(shape=(self.A,))
        self.R2X_k_cum = np.zeros(shape=(self.K, self.A))
        self.R2Y_m_cum = np.zeros(shape=(self.M, self.A))
        self.squared_prediction_error = np.zeros((self.N, self.A))

        # Perform MD algorithm here
        if self.missing_data_settings["md_method"].lower() == "pmp":
            self._fit_pmp_pls(X)
        elif self.missing_data_settings["md_method"].lower() in ["scp", "nipals"]:
            self._fit_nipals_pls(settings=self.missing_data_settings)
        elif self.missing_data_settings["md_method"].lower() in ["tsr"]:
            self._fit_tsr_pls(settings=self.missing_data_settings)

        # Additional calculations, which can be done after the missing data method is complete.
        # self.explained_variance_ = np.diag(self.x_scores.T @ self.x_scores) / (self.N - 1)
        return self

    def _fit_nipals_pls(self, settings):
        """
        Internal method to fit the PLS model using the NIPALS algorithm.
        """
        # NIPALS algorithm
        A = self.A

        # Initialize storage:
        self.extra_info = {}
        self.extra_info["timing"] = np.zeros(A) * np.nan
        self.extra_info["iterations"] = np.zeros(A) * np.nan

        for a in np.arange(A):

            # Timers and housekeeping
            start_time = time.time()
            itern = 0

            start_SSX_col = ssq(self.Xd, axis=0)
            start_SSY_col = ssq(self.Yd, axis=0)

            if sum(start_SSX_col) < epsqrt:
                emsg = (
                    "There is no variance left in the data array for X: cannot "
                    f"compute any more components beyond component {a}."
                )
                raise RuntimeError(emsg)
            if sum(start_SSY_col) < epsqrt:
                emsg = (
                    "There is no variance left in the data array for Y: cannot "
                    f"compute any more components beyond component {a}."
                )
                raise RuntimeError(emsg)

            # Initialize t_a with random numbers, or carefully select a column from X or Y?
            # Find a column with the largest variance as t1_start; replace missing with zeros
            # All columns have the same variance if the data have been scaled to unit variance!
            u_a_guess = self.Yd[:, [0]]
            u_a_guess[np.isnan(u_a_guess)] = 0
            u_a = u_a_guess + 1.0

            while not (terminate_check(u_a_guess, u_a, iterations=itern, settings=settings)):

                # 0: starting point for convergence checking on next loop
                u_a_guess = u_a.copy()

                # 1: Regress the score, u_a, onto every column in X, compute the
                #    regression coefficient and store in w_a
                # w_a = X.T * u_a / (u_a.T * u_a)
                w_a = quick_regress(self.Xd, u_a)

                # 2: Normalize w_a to unit length
                w_a /= np.sqrt(ssq(w_a))

                # 3: Now regress each row in X on the w_a vector, and store the
                #    regression coefficient in t_a
                # t_a = X * w_a / (w_a.T * w_a)
                t_a = quick_regress(self.Xd, w_a)

                # 4: Now regress score, t_a, onto every column in Y, compute the
                #    regression coefficient and store in c_a
                # c_a = Y * t_a / (t_a.T * t_a)
                c_a = quick_regress(self.Yd, t_a)

                # 5: Now regress each row in Y on the c_a vector, and store the
                #    regression coefficient in u_a
                # u_a = Y * c_a / (c_a.T * c_a)
                #
                # TODO(KGD):  % Still handle case when entire row in Y is missing
                u_a = quick_regress(self.Yd, c_a)

                itern += 1

            self.extra_info["timing"][a] = time.time() - start_time
            self.extra_info["iterations"][a] = itern

            if itern > settings["md_max_iter"]:
                Warning("PLS missing data [SCP method]: maximum number of iterations reached!")

            # Loop terminated!
            # 6: Now deflate the X-matrix.  To do that we need to calculate loadings for the
            # X-space.  Regress columns of t_a onto each column in X and calculate loadings, p_a.
            # Use this p_a to deflate afterwards.
            p_a = quick_regress(self.Xd, t_a)  # Note the similarity with step 4!
            self.Xd -= np.dot(t_a, p_a.T)  # and that similarity helps understand
            self.Yd -= np.dot(t_a, c_a.T)  # the deflation process.

            ## VIP value (only calculated for X-blocks); only last column is useful
            # self.stats.VIP_a = np.zeros((self.K, self.A))
            # self.stats.VIP = np.zeros(self.K)

            # Store results
            # -------------
            # Flip the signs of the column vectors in P so that the largest
            # magnitude element is positive (Wold, Esbensen, Geladi, PCA,
            # CILS, 1987, p 42)
            max_el_idx = np.argmax(np.abs(p_a))
            if np.sign(p_a[max_el_idx]) < 1:
                t_a *= -1.0
                u_a *= -1.0
                w_a *= -1.0
                p_a *= -1.0
                c_a *= -1.0

            # Store the loadings and scores
            self.x_scores_[:, a] = t_a.flatten()
            self.y_scores_[:, a] = u_a.flatten()
            self.x_weights_[:, a] = w_a.flatten()
            self.x_loadings_[:, a] = p_a.flatten()
            self.y_loadings_[:, a] = c_a.flatten()
            # end looping on ``a``


def ssq(X: np.ndarray, axis: Optional[int] = None) -> Any:
    """A function than calculates the sum of squares of a 2D matrix
    (not array! and not checked for either: code will simply fail),
    skipping over any NaN (missing) data.

    """
    N, K = X.shape
    if axis == 0:
        out_ax0 = np.zeros(K)
        for k in np.arange(K):
            out_ax0[k] += np.nansum(X[:, k] ** 2)

        return out_ax0

    if axis == 1:
        out_ax1 = np.zeros(N)
        for n in np.arange(N):
            out_ax1[n] += np.nansum(X[n, :] ** 2)

        return out_ax1

    out = 0.0
    if axis is None:
        out = np.nansum(X ** 2)

    return out


def terminate_check(
    t_a_guess: np.ndarray, t_a: np.ndarray, iterations: int, settings: dict
) -> bool:
    """The PCA iterative algorithm is terminated when any one of these conditions is True:

    #. scores converge: the norm between two successive iterations
    #. maximum number of iterations is reached
    """
    score_tol = np.linalg.norm(t_a_guess - t_a, ord=None)
    converged = score_tol < settings["md_tol"]
    max_iter = iterations > settings["md_max_iter"]
    if np.any([max_iter, converged]):
        return True
    else:
        return False


def quick_regress(Y, x):
    """
    Regress vector `x` onto the columns in matrix ``Y`` one at a time.
    Return the vector of regression coefficients, one for each column in `Y`.
    There may be missing data in `Y`, but not in `x`.  The `x` vector
    *must* be a column vector.
    """
    Ny, K = Y.shape
    Nx = x.shape[0]
    if Ny == Nx:  # Case A: b' = (x'Y)/(x'x): (1xN)(NxK) = (1xK)
        b = np.zeros((K, 1))
        for k in np.arange(K):
            b[k] = np.sum(x.T * np.nan_to_num(Y[:, k]))
            temp = ~np.isnan(Y[:, k]) * x.T
            denom = np.dot(temp, temp.T)[0][0]
            if np.abs(denom) > epsqrt:
                b[k] /= denom
        return b

    elif K == Nx:  # Case B: b = (Yx)/(x'x): (NxK)(Kx1) = (Nx1)
        b = np.zeros((Ny, 1))
        for n in np.arange(Ny):
            b[n] = np.sum(x[:, 0] * np.nan_to_num(Y[n, :]))
            # TODO(KGD): check: this denom is usually(always?) equal to 1.0
            denom = ssq(~np.isnan(Y[n, :]) * x.T)
            if np.abs(denom) > epsqrt:
                b[n] /= denom
        return b


def center(X, func=np.mean, axis=0, extra_output=False):
    """
    Performs centering of data, using a function, `func` (default: np.mean).
    The function, if supplied, but return a vector with as many columns as the matrix X.

    `axis` [optional; default=0] {integer or None}

    This specifies the axis along which the centering vector will be calculated if not provided.
    The function is applied along the `axis`: 0=down the columns; 1 = across the rows.

    *Missing values*: The sample mean is computed by taking the sum along the `axis`, skipping
    any missing data, and dividing by N = number of values which are present. Values which were
    missing before, are left as missing after.
    """
    vector = pd.DataFrame(X).apply(func, axis=axis).values
    if extra_output:
        return np.subtract(X, vector), vector
    else:
        return np.subtract(X, vector)


def scale(X, func=np.std, axis=0, extra_output=False, **kwargs):
    """
    Scales the data (does NOT do any centering); scales to unit variance by
    default.


    `func` [optional; default=np.std] {a function}
        The default (np.std) will use NumPy to calculate the sample standard
        deviation of the data, and use that as `scale`.

        TODO: provide a scaling vector.
        The sample standard deviation is computed along the required `axis`,
        skipping over any missing data, and dividing by N-1, where N = number
        of values which are present, i.e. not counting missing values.

    `axis` [optional; default=0] {integer or None}
        Transformations are applied on slices of data.  This specifies the
        axis along which the transformation will be applied.

    #`markers` [optional; default=None]
    #    A vector (or slice) used to store indices (the markers) where the
    ##    variance needs to be replaced with the `low_variance_replacement`
    #
    #`variance_tolerance` [optional; default=1E-7] {floating point}
    #    A slice is considered to have no variance when the actual variance of
    #    that slice is smaller than this value.
    #
    #`low_variance_replacement` [optional; default=0.0] {floating point}
    #    Used to replace values in the output where the `markers` indicates no
    #    or low variance.

    Usage
    =====

    X = ...  # data matrix
    X = scale(center(X))
    my_scale = np.mad
    X = scale(center(X), func=my_scale)

    """
    # options = {}
    # options["markers"] = None
    # options["variance_tolerance"] = epsqrt
    # options["low_variance_replacement"] = np.NaN

    vector = pd.DataFrame(X).apply(func, axis=axis, **kwargs).values
    # if options["markers"] is None:
    # options["markers"] = vector < options["variance_tolerance"]
    # if options["markers"].any():
    #    options["scale"][options["markers"]] = options["low_variance_replacement"]

    vector = 1.0 / vector

    if extra_output:
        return np.multiply(X, vector), vector
    else:
        return np.multiply(X, vector)


def T2_limit(conf_level: float = 0.95, n_components: int = 0, n_rows: int = 0) -> float:
    """Returns the Hotelling's T2 value at the given level of confidence.

    Parameters
    ----------
    conf_level : float, optional
        Fractional confidence limit, less that 1.00; by default 0.95

    Returns
    -------
    float
        The Hotelling's T2 limit at the given level of confidence.
    """
    assert conf_level > 0.0
    assert conf_level < 1.0
    assert n_rows > 0
    A, N = n_components, n_rows
    return A * (N - 1) * (N + 1) / (N * (N - A)) * f.isf((1 - conf_level), A, N - A)


def ellipse_coordinates(
    score_horiz: int,
    score_vert: int,
    T2_limit_conf_level: float = 0.95,
    n_points: int = 100,
    n_components: int = 0,
    scaling_factor_for_scores=None,
    n_rows: int = 0,
) -> tuple:
    """Get the (score_horiz, score_vert) coordinate pairs that form the T2 ellipse when
        plotting the score `score_horiz` on the horizontal axis and `score_vert` on the
        vertical axis.

        Scores are referred to by number, starting at 1 and ending with `model.A`


    Parameters
    ----------
    score_horiz : int
        [description]
    score_vert : int
        [description]
    T2_limit_conf_level : float
        The `conf_level` confidence value: e.g. 0.95 is for the 95% confidence limit.
    n_points : int, optional
        Number of points to use in the ellipse; by default 100.

    Returns
    -------
    tuple of 2 elements; the first for the x-axis; the second for the y-axis.
        Returns `n_points` equispaced points that can be used to plot an ellipse.

    Background
    ----------

    Equation of ellipse in *canonical* form (http://en.wikipedia.org/wiki/Ellipse)

        (t_horiz/s_h)^2 + (t_vert/s_v)^2  =  T2_limit_alpha
        s_horiz = stddev(T_horiz)
        s_vert  = stddev(T_vert)
        T2_limit_alpha = T2 confidence limit at a given alpha value

    Equation of ellipse, *parametric* form (http://en.wikipedia.org/wiki/Ellipse):

        t_horiz = sqrt(T2_limit_alpha)*s_h*cos(t)
        t_vert  = sqrt(T2_limit_alpha)*s_v*sin(t)

        where t ranges between 0 and 2*pi.
    """
    assert score_horiz >= 1
    assert score_vert >= 1
    assert score_horiz <= n_components
    assert score_vert <= n_components
    assert T2_limit_conf_level > 0
    assert T2_limit_conf_level < 1
    assert n_rows > 0
    s_h = scaling_factor_for_scores[score_horiz - 1]
    s_v = scaling_factor_for_scores[score_vert - 1]
    T2_limit_specific = np.sqrt(
        T2_limit(T2_limit_conf_level, n_components=n_components, n_rows=n_rows)
    )
    dt = 2 * np.pi / (n_points - 1)
    steps = np.linspace(0, n_points - 1, n_points)
    x = np.cos(steps * dt) * T2_limit_specific * s_h
    y = np.sin(steps * dt) * T2_limit_specific * s_v
    return x, y


# def _apply_pca(self, new=None):
#     """
#     Project new observations, ``new``, onto the existing latent variable
#     model.  Returns a ``Projection`` object, which contains the scores,
#     SPE, T2, and predictions.

#     If ``new`` is not provided it will return a ``Projection`` object
#     for the training data set.
#     """
#     # TODO: complete this code.
#     new = preprocess(LVM, new)
#     result = Projection()
#     result.scores = np.zeros([LVM.J, LVM.A])
#     result.SPE = np.zeros([LVM.J, 1])
#     result.T2 = np.zeros([LVM.J, 1])
#     result.Yhat = np.zeros([LVM.J, LVM.M])

#     # (1, 2, ... J) * every tag * for J time steps * for every LV
#     # result.c_scores = np.zeros([LVM.J, LVM.K, LVM.J, LVM.A]) * np.nan

#     K = LVM.K
#     x_project = new.copy()
#     for j in np.arange(LVM.J):
#         idx_beg, idx_end = K * j, K * (j + 1)
#         x = x_project.copy()
#         if LVM.opt.md_method == "scp":
#             for a in np.arange(LVM.A):

#                 if LVM.M:  # PLS
#                     r = LVM.W[0:idx_end, a]
#                 else:  # PCA
#                     r = LVM.P[0:idx_end, a]

#                 rtr = np.dot(r.T, r)
#                 p = LVM.P[0:idx_end, a]

#                 # t_{\text{new},j}(a) = \mathbf{P/W}'_*(:,a) \mathbf{x}_\text{pp}/DEN
#                 temp = (r * x[0, 0:idx_end]) / (rtr + 0.0)
#                 # result.c_scores[0:j+1, :, j, a] = temp.reshape(j+1, LVM.K)
#                 result.scores[j, a] = np.nansum(temp)

#                 # \mathbf{x}_\text{pp} -= t_{\text{new},j}(a) \mathbf{P}'_*(:,a)
#                 x[0, 0:idx_end] -= result.scores[j, a] * p

#             # The error_j is only the portion related to the current time step
#             error_j = x[0, idx_beg:idx_end]

#         result.SPE[j, 0] = np.nansum(error_j ** 2)
#         result.T2[j, 0] = np.sum((result.scores[j, :] / LVM.S[j, :]) ** 2)
#         Yhat_MCUV = np.dot(result.scores[j, :], LVM.C.T)
#         result.Yhat[j, :] = Yhat_MCUV / (LVM.PPY[1][1] + 0.0) + LVM.PPY[0][1]

#     return result



"""
IDEAS from earlier code written in 2015/2016.
RSM optimization.



    function out = optimization_caller()

% Leave this part intact. Copy it, and paste it below, altering the values.
% Repeat that once per factor.
template = struct;
template.descriptive_name = 'Parameter name';
template.argument_name = 'Name as used in a function';
template.parameter_type = 'Categorical | Numerical';
template.default_value = 'Text or numeric value';
template.levels = 'A cell array of level names, only for categorical factor. Numerical: ignored';
template.level_mapping = 'A vector of numeric values that are used to map the levels, e.g. [1, 2, 3]. Numerical: ignored';
template.starting_lower_bound = 'A numeric factor will be evaluated starting at this low value. Categorical: ignored';
template.starting_upper_bound = 'A numeric factor will be evaluated starting at this upper value, Categorical: ignored';
template.absolute_lower_bound = 'A numeric factor cannot go lower than this. Categorical: ignored';
template.absolute_upper_bound = 'A numeric factor cannot go above this. Categorical: ignored';


% 0. OVERALL settings
% ========================
settings = struct;

% This is your function that will be called to run the optimization.
% It will have only 1 input, a structure. That structure will have the names
% of your factors. If you have specified ``factors`` (see below) as
%   factors.A
%   factors.my_variable
%   etc
%
% that is what will be passed into the function.
%
% This will almost certainly mean that you must write another MATLAB file
% function which receives this input, and translates it into the right form to
% actually run the experiment/simulation.
%
% The function must return a single scalar value.
settings.expt_function = @my_function_call;
settings.expt_function__n_outputs = 3;

% Number of starting runs to kick-off the D-optimal design
settings.n_start_runs = 20;

% If you start with a full factorial, then the above setting of
% ``n_start_runs`` is ignored. It will run the full set of potentially many,
% many runs
settings.start_full_factorial = 0;



% 1. SPECIFY parameters
% ========================
% Reference spectrum type
A = template;
A.descriptive_name = 'Reference type';
A.argument_name = 'reference_type';
A.parameter_type = 'Categorical';
A.default_value = 'average';
A.levels = {'average', 'median', 'max', 'average2'};
A.level_mapping = [1, 2, 3, 4];

% Number of intervals
B = template;
B.descriptive_name = 'Number of intervals';
B.argument_name = 'number_intervals';
B.parameter_type = 'Numerical';
B.default_value = 50;
B.starting_lower_bound = 60;
B.starting_upper_bound = 120;
B.absolute_lower_bound = 10;
B.absolute_upper_bound = 100;

% Number of intervals: as a categorical variable
B = template;
B.descriptive_name = 'Number of intervals';
B.argument_name = 'number_intervals';
B.parameter_type = 'Categorical';
B.default_value = 50;
B.levels = {50, 70, 90, 110};
B.level_mapping = [50, 70, 90, 110];


% Filling mode
C = template;
C.descriptive_name = 'Filling mode';
C.argument_name = 'filling_mode';
C.parameter_type = 'Categorical';
C.default_value = 0;
C.levels = {0, 1};
C.level_mapping = [0, 1];


% Coshift preprocessing
D = template;
D.descriptive_name = 'Coshift';
D.argument_name = 'coshift_preprocessing';
D.parameter_type = 'Categorical';
D.default_value = 0;
D.levels = {0, 1};
D.level_mapping = [0, 1];

% Max allowed shift for the Co-shift preprocessing
E = template;
E.descriptive_name = 'Maximum allowed shift';
E.argument_name = 'max_coshift_allowed';
E.parameter_type = 'Numerical';
E.default_value = 50;
E.starting_lower_bound = 40;
E.starting_upper_bound = 100;
E.absolute_lower_bound = 10;
E.absolute_upper_bound = 100;

% Alignment method
F = template;
F.descriptive_name = 'Alignment';
F.argument_name = 'alignment_method';
F.parameter_type = 'Categorical';
F.default_value = 0;
F.levels = {0, 1, 3};
F.level_mapping = [0, 1, 3];


% Maximum shift correction in data
G = template;
G.descriptive_name = 'Maxshift';
G.argument_name = 'maxshift_type';
G.parameter_type = 'Categorical';
G.default_value = 'b';
G.levels = {'b', 'f'};
G.level_mapping = [0, 1];


factors = struct;
factors.A = A;
factors.B = B;
factors.C = C;
factors.D = D;
%factors.E = E;  % <-- not used anyway
factors.F = F;
%factors.G = G;

fields = fieldnames(factors);
n_factors = numel(fields);

% 2. CALCULATE all combinations
% ========================
% Also, while do this, call the ``cordexch`` function, to see what that gives:


type_of_experiment = 'interaction';


starting_set = cell(1, numel(fields));
categorical = zeros(1, n_factors);
levels = zeros(1, n_factors);
for p = 1 : n_factors
    factor = factors.(fields{p});
    if strcmp(factor.parameter_type, 'Numerical')
        this_levels = [factor.starting_lower_bound, factor.starting_upper_bound];
        factors.(fields{p}).current_mins_one = factor.starting_lower_bound;
        factors.(fields{p}).current_plus_one = factor.starting_upper_bound;
        levels(p) = 2;
    elseif strcmp(factor.parameter_type, 'Categorical')
        this_levels = factor.level_mapping;
        levels(p) = numel(factor.level_mapping);
        categorical(p) = 1;
    end
    starting_set{p} = this_levels;
end

% Based on code from https://nl.mathworks.com/matlabcentral/fileexchange/10064-allcomb-varargin
% but modified for our situation.
ii = 1 : 1 : numel(fields);
A = zeros(0, numel(fields)) ;
[A{ii}] = ndgrid(starting_set{ii});
combinations = reshape(cat(numel(fields) + 1, A{:}), [], numel(fields));

% Coordinate exchange experiments:
coord_exch_full_set = cordexch(n_factors, size(combinations, 1), type_of_experiment, ...
         'categorical', find(categorical), 'levels', levels, 'display', 'off');



% 3. SELECT the subset to run
% =========================

% TODO: select a Plackett-Burham subset

% Use the D-optimal approach to select a subset of the full set in
% ``combinations`` as a starting candidate set.

if settings.start_full_factorial
    start_set = combinations;

else
    start_set = rowexch(n_factors, settings.n_start_runs, type_of_experiment, ...
         'categorical', find(categorical), 'levels', levels, 'display', 'off');

    % Original method: use "candexch" to select, but this ignores all the
    % levels of a categorical factor, and only picks the extremes (first and
    % last levels of such a factor).
    %subset = candexch(coord_exch_full_set, settings.n_start_runs, 'display', 'off', ...
    %              'tries', 100, 'maxiter', 1000);
    %start_set = combinations(subset, :);
end

% Remove the duplicates
start_set = unique(start_set, 'rows', 'stable');


% 4. RUN the experiments
n_runs = size(start_set, 1);
outputs = zeros(n_runs, settings.expt_function__n_outputs);
for k = 1:n_runs
    disp(['Run number: ', num2str(k), ' of ', num2str(n_runs)])
    fn_input = translate_coded_to_realworld_structure(start_set(k, :), factors);
    fn_input
    %try
    outputs(k, :) = settings.expt_function(fn_input);
    %catch
    %
    %    outputs(k, :) = NaN;
    %end
    outputs(k, :)
end
out = [start_set outputs];

% 5. SHOW results
%------------------------------
figure;
subplot(settings.expt_function__n_outputs, n_factors, 1)
count = 1;
for yout = 1:settings.expt_function__n_outputs
    for p = 1:n_factors
        factor = factors.(fields{p});
        subplot(settings.expt_function__n_outputs, n_factors, count)
        count = count + 1;
        xs = start_set(:, p);
        ys = outputs(:, yout);
        plot(xs, ys, 'o')
        hold on
        b = regress(ys, [ones(size(xs, 1), 1), xs]);
        plot(xs, b(1) + b(2) .* xs, 'k-', 'linewidth', 2)
        grid on
        title(factor.descriptive_name)
    end

    fprintf('For output %d: the highest and lowest values are:\n', yout)
    [min_y, idx_min] = min(ys);
    fprintf('Low : y = %f; settings=\n', min_y)
    disp(translate_coded_to_realworld_structure(start_set(idx_min, :), factors))


    [max_y, idx_max] = max(ys);
    fprintf('High: y = %f; settings=\n', max_y)
    disp(translate_coded_to_realworld_structure(start_set(idx_max, :), factors))
end

a=2;


% 6. IDENTIFY next optimization
%------------------------------

function output = translate_coded_to_realworld_structure(input, factors)
% Translates a numeric vector, ``input``, where each entry is the numeric
% factor settings [which may include categorical values], back to the mixed
% numeric and string representation. Returns a structure.
%   range = upper_bound - lower_bound
%   center_value = 0.5 * (upper_bound + lower_bound)
%   coded_value = (actual_value - center_value) / (0.5 * range)
%   actual_value = coded_value * (0.5 * range) + lower_bound
fields = fieldnames(factors);
output = struct;
for p = 1 : numel(fields)
    factor = factors.(fields{p});
    if strcmp(factor.parameter_type, 'Numerical')

        % To get exact matching, without roundoff
        if input(p) == -1
            value = factor.current_mins_one;
        elseif input(p) == +1
            value = factor.current_plus_one;
        else
            range = (factor.current_plus_one - factor.current_mins_one);
            center_value = 0.5*(factor.current_plus_one + factor.current_mins_one);
            value = input(p) * (0.5 * range) + center_value;
        end

    elseif strcmp(factor.parameter_type, 'Categorical')
        value = factor.levels{input(p)};
        % Do NOT remove the 'find' part here, or try to simplify it. It is
        % correct as it is.
        %temp = %(factor.levels(find(factor.level_mapping == input(p))));
        %value = temp{1};
    end
    output.(factor.argument_name) = value;
end
"""



# (c) Kevin Dunn, 2010-2021. MIT License. Based on own private work over the years.

from numpy.random import normal
import pandas as pd


def popcorn(t=120, T=None):
    """
    Simulation of stovetop popcorn cooking.

    A fixed number of popcorn kernels are cooked are heated at the same
    temperature. This simulation returns the number of kernels that are edible
    when the pot is left on the stove for a given number of `t` seconds.


    Parameters
    ----------

    `t` is the number of seconds that the pot is left on the stove. The
    default amount, if not provided, is 120 seconds.

    Time durations less than 77 seconds are not supported. A vector (list) of
    time values is not permitted, since the goal is to perform sequential
    experimentation to determine the optimum time, with the fewest number of
    function calls.

    Returns
    -------
    The number of edible popcorn kernels. Random noise is added for realism.

    Source
    ------
    Kevin Dunn, Process Improvement using Data, Chapter 5, 2010 to 2021,
    https://learnche.org/pid


    Examples
    --------
    >>> popcorn(t=55)    # will fail
    >>> popcorn(t=120)
    >>> popcorn(t=152.2)

    >>> # What happens if we leave the pot on the stove for too long?
    >>> popcorn(t=500)


    Also see
    grocery
    manufacter

    """
    pass


def grocery(p=3.46, h=150, P=None, H=None):
    """
    Simulation of grocery store profits for a single product.

    The hourly profit made when selling the product at price `p` and the product
    is displayed at height `h` [cm up from the ground] on the shelf.

    Simulates a grocery store profit function where there are 2 factors:
    * `p` = selling price of the product, measured in dollars and cents
    * `h` = height of the product on the shelf, measured in centimeters above
          the ground.

    Typical values are p = $3.50 and h = 150cm
    The outcome is: profit made per hour [dollars/hour], with random noise
                    added, for realism.

    Source
    ------
    Kevin Dunn, Process Improvement using Data, Chapter 5, 2010 to 2021,
    https://learnche.org/pid

    """
    if P is None:
        P = p
    if H is None:
        H = h

    if (len(P) > 1) | (len(H) > 1):
        assert (
            False
        ), "Running the grocery store experiments in parallel is (intentionally) not allowed."

    if pd.isna(P) or pd.isna(H):
        assert False, "All function inputs must be finite numbers."
    elif P < 0:
        assert False, "Please provide a positive sales price, P."
    elif H < 0:
        assert False, "The height of the shelving, H, must be a positive value."

    a_coded = (P - 3.2) / 0.2
    b_coded = (H - 50) / 100
    y = round(
        (
            18 * a_coded
            + 12 * b_coded
            - 7 * a_coded * a_coded
            - 6.0 * b_coded * b_coded
            - 8.5 * a_coded * b_coded
            + 60
        )
        * 10.0
        + normal(0, 1) * 2
    )
    return y



from typing import Dict, List, Any, Union

import numpy as np
import pandas as pd
import scipy as sp
import plotly.graph_objects as go

from .alignment_helpers import distance_matrix, backtrack_optimal_path
from .data_input import dict_to_wide, check_valid_batch_dict, melted_to_dict
from ..multivariate.methods import MCUVScaler, PCA

epsqrt = np.sqrt(np.finfo(float).eps)


def determine_scaling(
    batches: Dict[str, pd.DataFrame],
    columns_to_align: List = None,
    settings: dict = None,
) -> pd.DataFrame:
    """
    Scales the batch data according to the variable ranges.

    Parameters
    ----------
    dict_df : Dict[str, pd.DataFrame]
        Batch data, in the standard format.

    Returns
    -------
    range_scalers: DataFrame

        J rows
        2 columns:
            column 1 = Range of each tag. Approximated as the typical delta from q98 - q02
            column 2 = Typical minimum of each tag. Robustly calculated.

    TODO: put this in a scikit-learn style: .fit() and .apply() style
    """
    # This will be clumsy, until we have Python 3.9
    default_settings = {"robust": True}
    if settings:
        default_settings.update(settings)

    settings = default_settings
    if columns_to_align is None:
        columns_to_align = batches[list(batches.keys())[0]].columns

    collector_rnge = []
    collector_mins = []
    for _, batch in batches.items():
        if settings["robust"]:
            # TODO: consider f_iqr feature here. Would that work?
            rnge = batch[columns_to_align].quantile(0.98) - batch[columns_to_align].quantile(0.02)
        else:
            rnge = batch[columns_to_align].max() - batch[columns_to_align].min()

        rnge[rnge.values == 0] = 1.0
        collector_rnge.append(rnge)
        collector_mins.append(batch.min(axis=0))

    if settings["robust"]:
        scalings = pd.concat(
            [
                pd.DataFrame(collector_rnge).median(),
                pd.DataFrame(collector_mins).median(),
            ],
            axis=1,
        )
    else:
        scalings = pd.concat(
            [pd.DataFrame(collector_rnge).mean(), pd.DataFrame(collector_mins).mean()],
            axis=1,
        )
    scalings.columns = ["Range", "Minimum"]
    return scalings


def apply_scaling(
    batches: Dict[str, pd.DataFrame],
    scale_df: pd.DataFrame,
    columns_to_align: List = None,
) -> dict:
    """Scales the batches according to the information in the scaling dataframe.

    Parameters
    ----------
    batches : Dict[str, pd.DataFrame]
        The batches, in standard format.
    scale_df : pd.DataFrame
        The scaling dataframe, from `determine_scaling`

    Returns
    -------
    dict
        The scaled batch data.
    """
    # TODO: handle the case of DataFrames still
    if columns_to_align is None:
        if isinstance(batches, dict):
            batch1 = batches[list(batches.keys())[0]]
            columns_to_align = batch1.columns
        elif isinstance(batches, pd.DataFrame):
            columns_to_align = batches.columns
        else:
            assert False, "Undefined input type"
    out = {}
    for batch_id, batch in batches.items():
        out[batch_id] = batch[columns_to_align].copy()
        for tag, column in out[batch_id].iteritems():
            out[batch_id][tag] = (column - scale_df.loc[tag, "Minimum"]) / scale_df.loc[
                tag, "Range"
            ]
    return out


def reverse_scaling(
    batches: Dict[str, pd.DataFrame],
    scale_df: pd.DataFrame,
    columns_to_align: List = None,
):
    # TODO: handle the case of DataFrames still
    if columns_to_align is None:
        if isinstance(batches, dict):
            columns_to_align = batches[list(batches.keys())[0]].columns
        elif isinstance(batches, pd.DataFrame):
            columns_to_align = batches.columns
        else:
            assert False, "Undefined input type"
    out = {}
    for batch_id, batch in batches.items():
        out[batch_id] = batch[columns_to_align].copy()
        for tag, column in out[batch_id].iteritems():
            out[batch_id][tag] = column * scale_df.loc[tag, "Range"] + scale_df.loc[tag, "Minimum"]
    return out


class DTWresult:
    """Result class."""

    def __init__(
        self,
        synced: np.ndarray,
        penalty_matrix: np.ndarray,
        md_path: np.ndarray,  # multi-dimensional path through distance mesh D = penalty_matrix
        warping_path: np.ndarray,
        distance: float,
        normalized_distance: float,
    ):
        self.synced = synced
        self.penalty_matrix = penalty_matrix
        self.md_path = md_path
        self.warping_path = warping_path
        self.distance = distance
        self.normalized_distance = normalized_distance


def align_with_path(md_path, batch, initial_row):
    row = 0
    nr = md_path[:, 0].max() + 1  # to account for the zero-based indexing
    synced = pd.DataFrame(np.zeros((nr, batch.shape[1])), columns=batch.columns)
    synced.iloc[row, :] = batch.iloc[md_path[0, 1], :]
    temp = initial_row
    for idx in np.arange(1, md_path.shape[0]):
        if md_path[idx, 0] != md_path[idx - 1, 0]:
            row += 1
            synced.iloc[row, :] = temp = batch.iloc[md_path[idx, 1], :]

        else:
            # TODO : Come back to page 181 of thesis: where more than 1 point in the target
            #        trajectory is aligned with the reference: compute the average,
            temp = np.vstack((temp, batch.iloc[md_path[idx, 1], :]))
            synced.iloc[row, :] = np.nanmean(temp, axis=0)

    return pd.DataFrame(
        synced,
    )


def dtw_core(test, ref, weight_matrix: np.ndarray):

    show_plot = False
    nt = test.shape[0]  # 'test' data; will be align to the 'reference' data
    nr = ref.shape[0]
    assert test.shape[1] == ref.shape[1]

    D = distance_matrix(test.values, ref.values, weight_matrix)
    md_path, distance = backtrack_optimal_path(D)
    warping_path = np.zeros(nr)
    for idx in range(nr):
        warping_path[idx] = md_path[np.where(md_path[:, 0] == idx)[0][-1], 1]

    # Now align the `test` batch:
    initial_row = ref.iloc[md_path[0, 0], :].copy()
    synced = align_with_path(md_path=md_path, batch=test, initial_row=initial_row)

    if show_plot:  # for debugging
        X = np.arange(0, nt, 1)
        Y = np.arange(0, nr, 1)
        X, Y = np.meshgrid(X, Y)
        fig = go.Figure(
            data=[
                go.Mesh3d(
                    x=X.ravel(),
                    y=Y.ravel(),
                    z=D.ravel(),
                    color="lightpink",
                    opacity=0.90,
                )
            ]
        )
        fig.show()

    return DTWresult(
        synced,
        D,
        md_path,
        warping_path,
        distance,
        normalized_distance=distance / (nr + nt),
    )


def one_iteration_dtw(
    batches_scaled: dict,
    refbatch_sc: pd.DataFrame,
    weight_matrix: np.ndarray,
    settings: dict = None,
):
    default_settings = {"show_progress": True, "subsample": 1}
    if settings:
        default_settings.update(settings)
    settings = default_settings

    aligned_batches = {}
    distances = []
    average_batch = refbatch_sc.copy().reset_index(drop=True) * 0.0
    successful_alignments = 0
    for batch_id, batch in batches_scaled.items():
        try:
            # see Kassidas, page 180
            batch_subset = batch.iloc[:: int(settings["subsample"]), :]
            result = dtw_core(batch_subset, refbatch_sc, weight_matrix=weight_matrix)
            average_batch += result.synced
            aligned_batches[batch_id] = result
            successful_alignments += 1
            distances.append(
                {
                    "batch_id": batch_id,
                    "Distance": result.distance,
                    "Normalized distance": result.normalized_distance,
                }
            )
            if settings["show_progress"]:
                message = f"  * {batch_id}: distance = {result.distance}"
                print(message)
        except ValueError:
            assert False, f"Failed on batch {batch_id}"

    average_batch /= successful_alignments

    return aligned_batches, average_batch


def batch_dtw(
    batches: Dict[str, pd.DataFrame],
    columns_to_align: list,
    reference_batch: str,
    settings: dict = None,
) -> dict:
    """
    Synchronize, via iterative DTW, with weighting.

    Algorithm: Kassidas et al. (2004):  https://doi.org/10.1002/aic.690440412

    Parameters
    ----------
    batches : Dict[str, pd.DataFrame]
        Batch data, in the standard format.
    columns_to_align : list
        Which columns to use during the alignment process. The others are aligned, but
        get no weight, and therefore do not influence the objective function.
    reference_batch : str
        Which key in the `batches` is the reference batch to use.
    settings : dict
        Default settings are = {
            "maximum_iterations": 25,  # maximum iterations (stops here, even if not converged)
            "tolerance": 0.1,  # convergence tolerance
            "robust": True,  # use robust scaling
            "show_progress": True,  # show progress
            "subsample": 1,  # use every sample
            "interpolate_time_axis_maximum": 100,  # interpolates everything to be on this scale
            "interpolate_time_axis_delta": 1,  # with this resolution
            "interpolate_method": "cubic" # any method from scipy.interpolate.interp1d allowed
        }

        The default settings will therefore resampled the time axis to have 100 data points,
        starting at 0 and ending at 99. You might want more points (change the delta), or use
        a different x-axis maximum.

    Returns
    -------
    dict
        Various outputs relevant to the alignment.
        TODO: Document completely later.

    Notation
    --------
    I = number of batches: index = i
    i = index for the batches
    J = number of tags (columns in each batch)
    j = index for the tags
    k = index into the rows of each batch, the samples: 0 ... k ... K_i
    """
    default_settings: Dict[str, Any] = dict(
        maximum_iterations=25,  # maximum iterations (stops here, even if not converged)
        tolerance=0.1,  # convergence tolerance
        robust=True,  # use robust scaling
        show_progress=True,  # show progress
        subsample=1,  # use every sample
        interpolate_time_axis_maximum=100,  # interpolates everything to be on this scale
        interpolate_time_axis_delta=1,
        interpolate_method="cubic",  # any method from scipy.interpolate.interp1d allowed
    )
    if settings:
        default_settings.update(settings)
    settings = default_settings
    assert settings["maximum_iterations"] >= 3, "At least 3 iterations are required"
    assert reference_batch in batches, "`reference_batch` was not found in the dict of batches."

    assert check_valid_batch_dict(batches, no_nan=True)

    scale_df = determine_scaling(
        batches=batches, columns_to_align=columns_to_align, settings=settings
    )
    batches_scaled = apply_scaling(batches, scale_df, columns_to_align)
    refbatch_sc = batches_scaled[reference_batch].iloc[:: int(settings["subsample"]), :]
    weight_vector = np.ones(refbatch_sc.shape[1])
    weight_matrix = np.diag(weight_vector)
    weight_history = np.zeros_like(weight_vector) * np.nan
    average_batch = None
    delta_weight = np.linalg.norm(weight_vector)
    iter = 0
    while (np.linalg.norm(delta_weight) > settings["tolerance"]) and (
        iter <= settings["maximum_iterations"]
    ):
        if settings["show_progress"]:
            message = f"Iter = {iter} and norm = {np.linalg.norm(delta_weight)}"
            print(message)

        iter += 1
        weight_matrix = np.diag(weight_vector)
        weight_history = np.vstack((weight_history, weight_vector.copy()))

        if iter > 3:
            refbatch_sc = average_batch

        aligned_batches, average_batch = one_iteration_dtw(
            batches_scaled=batches_scaled,
            refbatch_sc=refbatch_sc,
            weight_matrix=weight_matrix,
            settings=settings,
        )

        # Deviations from the average batch:
        next_weights = np.zeros((1, refbatch_sc.shape[1]))
        for batch_id, result in aligned_batches.items():
            next_weights += np.nansum(np.power(result.synced - average_batch, 2), axis=0)
            # TODO: use quadratic weights for now, but try sum of the absolute values instead
            #  np.abs(result.synced - average_batch).sum(axis=0)

            # TODO: leave out worst batches when computing the weights
            # dist_df = pd.DataFrame(distances).set_index("batch_id")
            # dist_df.hist("Distance", bins=50)
            # Now find the average trajectory, but ignore problematic batches:
            #      for example, top 5% of the distances.
            # TODO: make this a configurable setting
            # problematic_threshold = dist_df["Distance"].quantile(0.95)

        next_weights = 1.0 / np.where(next_weights > epsqrt, next_weights, 10000)
        weight_vector = (next_weights / np.sum(next_weights) * len(columns_to_align)).ravel()
        # If change in delta_weight is small, we terminate early; no need to fine-tune excessively.
        delta_weight = np.diag(weight_matrix) - weight_vector  # old - new

    # OK, the weights are found: now use the last iteration's result to get back to original
    # scaling for the trajectories
    weight_history = weight_history[1:, :]
    aligned_df = pd.DataFrame()
    new_time_axis = np.arange(
        0,
        settings["interpolate_time_axis_maximum"],
        settings["interpolate_time_axis_delta"],
    )

    for batch_id, result in aligned_batches.items():
        if settings["show_progress"]:
            message = f"Iterpolating values for batch = {batch_id}"
            print(message)

        initial_row = batches[batch_id].iloc[result.md_path[0, 0], :].copy()
        synced = align_with_path(
            result.md_path,
            batches[batch_id].iloc[:: int(settings["subsample"]), :],
            initial_row=initial_row,
        )
        # Resample the trajectories of the aligned data now along this sequence.
        sequence = np.linspace(
            0,
            settings["interpolate_time_axis_maximum"] - settings["interpolate_time_axis_delta"],
            synced.shape[0],
        )
        assert new_time_axis.min() == sequence.min()
        assert new_time_axis.max() == sequence.max()

        synced_interpolated = pd.DataFrame()
        for column, _ in synced.iteritems():
            if column in ["batch_id", "_sequence_"]:
                continue

            interp_column = sp.interpolate.interp1d(
                sequence,
                synced[column],
                kind=settings["interpolate_method"],
                assume_sorted=True,
            )
            synced_interpolated[column] = interp_column(new_time_axis)

        # Pop in an extra column at the start of the df
        synced_interpolated.insert(0, "batch_id", batch_id)
        synced_interpolated.set_index(new_time_axis)

        # Overwrite existing dataframe with this, unscaled, and interpolated dataframe.
        result.synced = synced_interpolated
        aligned_df = aligned_df.append(synced_interpolated)

    # Make the batch_id label consistent
    aligned_df["batch_id"] = aligned_df["batch_id"].astype(type(batch_id))

    last_average_batch = reverse_scaling(dict(avg=average_batch), scale_df)["avg"]
    aligned_batch_dfdict = melted_to_dict(aligned_df, batch_id_col="batch_id")

    return dict(
        scale_df=scale_df,
        aligned_batch_objects=aligned_batches,
        aligned_batch_dfdict=aligned_batch_dfdict,
        last_average_batch=last_average_batch,
        weight_history=pd.DataFrame(weight_history, columns=columns_to_align),
    )


def resample_to_reference(
    batches: Dict[str, pd.DataFrame],
    columns_to_align: list,
    reference_batch: str,
    settings: dict = None,
) -> dict:
    """Resamples all `batches` (only the `columns_to_align`) to the duration of batch with
    identifier `reference`.

    Parameters
    ----------
    batches : Dict[str, pd.DataFrame]
        Batch data, in the standard format.
    columns_to_align : list
        Which columns to use. Others are ignored.
    reference_batch : str
        Which key in the `batches` is the reference batch.
    settings : dict, optional
        [description], by default None

    Returns
    -------
    dict
        Batch data, in the standard format.
    """
    default_settings = {
        "interpolate_kind": "cubic",  # must be a valid "scipy.interpolate.interp1d" `kind`
    }
    if settings:
        default_settings.update(settings)
    settings = default_settings
    out = {}
    target_time = np.arange(0, batches[reference_batch].shape[0])
    target_time = target_time / target_time[-1]

    for batch_id, batch in batches.items():
        to_resample = np.arange(0, batch.shape[0])
        to_resample = to_resample / to_resample[-1]
        out_df = {}
        for column, series in batch.iteritems():
            if column in columns_to_align:
                out_df[column] = sp.interpolate.interp1d(
                    to_resample,
                    series.values,
                    copy=False,
                    kind=settings["interpolate_kind"],
                )(target_time)

        out[batch_id] = pd.DataFrame(out_df)

    return out


def find_average_length(batches: Dict[str, pd.DataFrame], settings: dict = None):
    """
    Find the batch in `batches` with the average length.

    Parameters
    ----------
    batches : Dict[str, pd.DataFrame]
        Batch data, in the standard format.
    settings : dict
        Default settings are = {
            "robust": True,  # use robust (median)
        }

    Returns
    -------
    One of the dictionary keys from `batches`.

    """
    default_settings = {
        "robust": True,  # use robust scaling
    }
    if settings:
        default_settings.update(settings)
    settings = default_settings

    batch_lengths = pd.Series({batch_id: df.shape[0] for batch_id, df in batches.items()})
    if settings["robust"]:
        # If multiple batches of the median length, return the last one.
        return batch_lengths.index[
            np.where((batch_lengths == batch_lengths.median()).values)[0][-1]
        ]
    else:
        return batch_lengths.index[(batch_lengths - batch_lengths.mean()).abs().argmin()]


def find_reference_batch(
    batches: Dict[str, pd.DataFrame],
    columns_to_align: list,
    settings: Dict[str, Any] = None,
):
    """
    Find a reference batch. Assumes NO missing data.

    Starts with the average duration batch; resamples (simple interpolation) of all batches to
    that duration. Unfolds that resampled data. Does PCA on the wide, unfolded data. Fits,
    by default, 4 components. Excludes all batches with Hotelling's T2 > 90% limit. Refits PCA
    with 4 components. Finds the batch which has the multivariate combination of scores which are
    the smallest (i.e. closest to the model center) and ensures this batch has SPE < 50% of the
    model limit.

    Parameters
    ----------
    batches : Dict[str, pd.DataFrame]
        Batch data, in the standard format.
    columns_to_align : list
        Which columns to use. Others are ignored.
    settings : dict
        Default settings are = {
            "robust": True,  # use robust scaling
            "subsample": 1,  # use every sample
            "method": "pca_most_average", # the most average batch from a crudely aligned PCA
            "n_components": 4,
        }

    Returns
    -------
    One of the dictionary keys from `batches`.

    """
    default_settings: Dict[str, Union[int, float, str, bool]] = {
        "robust": True,  # use robust scaling
        "subsample": 1,  # use every sample
        "method": "pca_most_average",
        "n_components": 4,
    }
    if isinstance(settings, dict):
        default_settings.update(settings)
    settings = default_settings

    assert check_valid_batch_dict(batches)

    # Starts with the average duration batch.
    initial_reference_id = find_average_length(batches, settings)

    # Resamples (simple interpolation) of all batches to that duration.
    resampled = resample_to_reference(
        batches,
        columns_to_align,
        reference_batch=initial_reference_id,
        settings=settings,
    )

    # Unfolds that resampled data.
    basewide = dict_to_wide(resampled)

    # Does PCA on the wide, unfolded data. A=4
    scaler = MCUVScaler().fit(basewide)
    mcuv = scaler.fit_transform(basewide)
    n_components = int(np.floor(settings["n_components"]))
    pca_first = PCA(n_components=n_components).fit(mcuv)

    # Excludes all batches with Hotelling's T2 > 90% limit.
    T2_limit_90 = pca_first.T2_limit(0.90)
    to_keep = pca_first.Hotellings_T2.iloc[:, -1] < T2_limit_90

    # Refits PCA with A=4 on a subset of the batches, to avoid biasing the PCA model too much.
    basewide = basewide.loc[to_keep, :]
    scaler = MCUVScaler().fit(basewide)
    mcuv = scaler.fit_transform(basewide)
    pca_second = PCA(n_components=n_components).fit(mcuv)

    # Finds batch with scores; and ensures this batch has SPE < 50% of the model limit.
    metrics = pd.DataFrame(
        {
            "HT2": pca_second.Hotellings_T2.iloc[:, -1],
            "SPE": pca_second.squared_prediction_error.iloc[:, -1],
        }
    )
    metrics = metrics.sort_values(by=["HT2", "SPE"])
    metrics = metrics.query(f"SPE < {pca_second.SPE_limit(0.5)}")
    return metrics.index[0]



"""
Functions in this files will ONLY use NumPy, and are therefore candidates for speed up with Numba.
"""

import numpy as np
from numba import jit


@jit(nopython=True)
def distance_matrix(test: np.ndarray, ref: np.ndarray, weight_matrix: np.ndarray):
    # TODO: allow user to specify `band`. The code below assumes that `band` is fixed as shown
    # here, so therefore, if user provide `band`, the code needs to be adjusted.
    nt = test.shape[0]  # 'test' data; will be align to the 'reference' data
    nr = ref.shape[0]
    band = np.ones((nt, 2), dtype=np.int64)
    band[:, 1] = nr
    dist = np.zeros((nr, nt))

    # Mahalanobis distance:
    for idx, row in enumerate(test):  # reset_index(drop=True).iterrows():
        dist[:, idx] = np.diag((row - ref) @ weight_matrix @ ((row - ref).T))

    # TODO: Sakoe-Chiba constraints could still be added
    D = np.zeros((nr, nt)) * np.NaN
    D[0, 0] = dist[0, 0]
    for idx in np.arange(1, nt):
        D[0, idx] = dist[0, idx] + D[0, idx - 1]

    for idx in np.arange(1, nr):
        D[idx, 0] = dist[idx, 0] + D[idx - 1, 0]

    for n in np.arange(1, nt):
        for m in np.arange(max((1, band[n, 0])), band[n, 1]):
            # index here must be integer!
            D[m, n] = dist[m, n] + np.nanmin(
                [D[m, n - 1], D[m - 1, n - 1], D[m - 1, n]]
            )

    return D


@jit(nopython=True)
def backtrack_optimal_path(D: np.ndarray):
    nr, nt = D.shape
    nr -= 1
    nt -= 1
    path_sum = 0.0
    path = [
        [nr, nt],
    ]
    while (nt + nr) != 0:
        if nt == 0:
            nr -= 1
            path_sum += D[nr, nt]
        elif nr == 0:
            nt -= 1
            path_sum += D[nr, nt]
        else:
            # Commented-code here is to read, but for Numba JIT, the other code is able to be
            # compiled. They give the same results in regular Python.
            # number = np.argmin([D[nr - 1, nt - 1], D[nr, nt - 1], D[nr - 1, nt]])
            a, b, c = D[nr - 1, nt - 1], D[nr, nt - 1], D[nr - 1, nt]
            if (a <= b) & (a <= c):
                # assert number == 0
                path_sum += D[nr - 1, nt - 1]
                nt -= 1
                nr -= 1
            elif (b <= a) & (b <= c):
                # assert number == 1
                path_sum += D[nr, nt - 1]
                nt -= 1
            elif (c <= a) & (c <= b):
                # assert number == 2
                path_sum += D[nr - 1, nt]
                nr -= 1
            else:
                assert False

        path.append([nr, nt])

    # All done:
    path.reverse()
    return np.array(path), path_sum



import pathlib
import unittest
import numpy as np
import pandas as pd


from process_improve.experiments.structures import gather, create_names, c
from process_improve.experiments.models import lm


class TestStructures(unittest.TestCase):
    """
    Test the data structures.
    """

    def setUp(self):
        self.A = [-1, +1, -1, +1, 0, +1]
        self.B = [-1, -1, +1, +1, 0, 0]
        self.C = [4, 5, 6, 4, 6]
        self.D = [0, 1, "green"]
        self.y = [52, 74, 62, 80, 50, 65]

    def test_create_names(self):
        """Creating factor names." """

        self.assertListEqual(create_names(5), ["A", "B", "C", "D", "E"])

        self.assertListEqual(create_names(3, letters=False), ["X1", "X2", "X3"])

        self.assertListEqual(
            create_names(3, letters=False, prefix="Q", start_at=9, padded=True),
            ["Q09", "Q10", "Q11"],
        )

        self.assertListEqual(
            create_names(4, letters=False, prefix="Z", start_at=99, padded=True),
            ["Z099", "Z100", "Z101", "Z102"],
        )

    def test_create_factors(self):
        A1 = c(*(self.A))
        A2 = c(*(self.A), index=["lo", "hi", "lo", "hi", "cp", "hi"])
        B = c(*(self.B), name="B")
        C1 = c(*(self.C), range=(4, 6))
        C2 = c(*(self.C), center=5, range=(4, 6))
        C3 = c(*(self.C), lo=4, hi=6)
        C4 = c(*(self.C), lo=4, hi=6, name="C4")
        C5 = c([4, 5, 6, 4, 6], lo=4, hi=6, name="C5")
        C6 = c(*(self.C), lo=5, hi=6, name="C6")
        y = c(*(self.y), name="conversion", units="%")

        self.assertTrue(isinstance(A1, pd.Series))
        self.assertTrue(A1.shape == (6,))
        self.assertTrue(hasattr(A1, "pi_index"))
        self.assertTrue(hasattr(A1, "name"))

        self.assertTrue(A1.name == "Unnamed [coded]")
        self.assertTrue(hasattr(A1, "pi_lo"))
        self.assertTrue(A1.pi_lo == -1)
        self.assertTrue(hasattr(A1, "pi_hi"))
        self.assertTrue(A1.pi_hi == +1)
        self.assertTrue(hasattr(A1, "pi_range"))
        self.assertTrue(A1.pi_range[0] == -1)
        self.assertTrue(A1.pi_range[1] == +1)
        self.assertTrue(hasattr(A1, "pi_center"))
        self.assertTrue(A1.pi_center == 0)

        self.assertTrue(isinstance(A2.index, pd.Index))
        self.assertTrue(hasattr(A2, "pi_index"))
        self.assertTrue(A2.name == "Unnamed [coded]")

        with self.assertRaises(IndexError):
            A2 = c(*(self.A), index=["lo", "hi", "lo", "hi", "cp"])

        self.assertTrue(B.shape == (6,))
        self.assertTrue(B.name == "B [coded]")

        self.assertTrue(C1.pi_range == (4, 6))
        self.assertTrue(C2.pi_center == 5)
        self.assertTrue(C2.pi_range == (4, 6))
        self.assertTrue(C3.pi_lo == 4)
        self.assertTrue(C3.pi_hi == 6)
        self.assertTrue(C4.pi_lo == 4)
        self.assertTrue(C4.pi_hi == 6)
        self.assertTrue(C5.pi_hi == 6)
        self.assertTrue(C5.name == "C5")

        # User says the low is 5, but the minimum is actually different
        self.assertTrue(C6.pi_lo == 5)
        self.assertTrue(C6.pi_range == (5, 6))

        D = c(*(self.D))
        self.assertTrue(D.pi_numeric is True)

        self.assertTrue(len(y) == 6)
        self.assertTrue(y.name == "conversion [%]")

    def test_column_math(self):
        self.A = [-1, +1, -1, +1, 0, +1]
        self.B = [-1, -1, +1, +1, 0, 0]
        A = c(*(self.A))
        B = c(*(self.B), name="B")
        C1 = A * B
        C2 = B * A

        self.assertTrue(np.all(C1.values == [+1, -1, -1, +1, 0, 0]))
        self.assertTrue(np.all(C1.index == A.index))
        self.assertTrue(np.all(C2.values == [+1, -1, -1, +1, 0, 0]))
        self.assertTrue(np.all(C2.index == A.index))

    def test_gather(self):
        A = c(*(self.A))
        B = c(*(self.B))
        y = c(*(self.y), name="conversion")

        expt = gather(A=A, B=B, y=y)
        self.assertTrue(expt.shape == (6, 3))

        expt = gather(A=A, B=B, y=y, title="Testing expt name")

        # Eventually this method must go to the "DF" class; currently in the
        # model class; not really appropriate there.
        self.assertTrue(expt.get_title() == "Testing expt name")


class TestModels(unittest.TestCase):
    def setUp(self):

        A = c(-1, +1, -1, +1)
        B = c(-1, -1, +1, +1)
        C = A * B
        y = c(41, 27, 35, 20, name="Stability", units="days")
        self.expt = gather(A=A, B=B, C=C, y=y, title="Half-fraction, using C = A*B")
        self.model_stability_poshalf = lm("y ~ A*B*C", self.expt)

        """
        Results from R:
        A = c(-1, +1, -1, +1)
        B = c(-1, -1, +1, +1)
        C = A*B
        y = c(41, 27, 35, 20)
        data = data.frame(A=A, B=B, C=C, y=y)
        model = lm(y ~ A*B*C, data=data)
        summary(model)

        Call:
            lm(formula = y ~ A * B * C, data = data)

            Residuals:
            ALL 4 residuals are 0: no residual degrees of freedom!

            Coefficients: (4 not defined because of singularities)
                        Estimate Std. Error t value Pr(>|t|)
            (Intercept)    30.75         NA      NA       NA
            A              -7.25         NA      NA       NA
            B              -3.25         NA      NA       NA
            C              -0.25         NA      NA       NA
            A:B               NA         NA      NA       NA
            A:C               NA         NA      NA       NA
            B:C               NA         NA      NA       NA
            A:B:C             NA         NA      NA       NA

            Residual standard error: NaN on 0 degrees of freedom
            Multiple R-squared:      1,	Adjusted R-squared:    NaN
            F-statistic:   NaN on 3 and 0 DF,  p-value: NA
        """

    def test_half_fraction(self):
        """
        Testing attributes for the half-fraction model.
        """
        self.assertTrue(self.model_stability_poshalf.nobs == 4)
        self.assertTrue(self.model_stability_poshalf.df_resid == 0)
        beta = self.model_stability_poshalf.get_parameters(drop_intercept=False)

        self.assertAlmostEqual(beta["A"], -7.25)
        self.assertAlmostEqual(beta["B"], -3.25)
        self.assertAlmostEqual(beta["C"], -0.25)
        self.assertAlmostEqual(beta["Intercept"], 30.75)
        self.assertTrue(
            self.model_stability_poshalf.get_aliases()
            == ["A + B:C", "B + A:C", "C + A:B"]
        )

        for resid in self.model_stability_poshalf.residuals:
            self.assertAlmostEqual(resid, 0.0)


class Test_API_usage(unittest.TestCase):
    def setUp(self):
        # TODO: replace unittest -> pytest
        # TODO: use path
        folder = (
            pathlib.Path(__file__).parents[1]
            / "process_improve"
            / "datasets"
            / "experiments"
        )
        self.df1 = pd.read_csv(folder / "test_doe1.csv")
        self.df1.set_index("Run order", inplace=True)

    def test_case_1(self):

        index = self.df1.index
        C = c(
            self.df1["C"],
            lo=self.df1["C"].min(),
            hi=self.df1["C"].max(),
            index=index,
            name="C",
        )
        M = c(self.df1["M"], levels=self.df1["M"].unique(), name="M")
        V = c(self.df1["V"], lo=self.df1["V"].min(), hi=self.df1["V"].max(), name="V")
        B = c(self.df1["B"], name="B")
        self.assertTrue(B.pi_levels["B"] == ["Ard", "Eme"])

        y = self.df1["y"]

        expt = gather(C=C, M=M, V=V, B=B, y=y)
        self.assertTrue(np.all(C.index == M.index))

        _ = lm("np.log10(y) ~ C*M*B*V", expt)
        # summary(model)
        # pareto_plot(model)
        # contour_plot(model, C, M)
        # contour_plot(model, "C", "M")
        # predict_plot(model)

    def test_aliasing(self):
        d4 = c(24, 48, 36, 36, 60, units="hours", lo=24, high=48)
        y4 = c(31, 65, 52, 54, 69)
        expt4 = gather(d=d4, y=y4, title="RW units")
        model4 = lm("y ~ d + I(np.power(d, 2))", data=expt4)
        self.assertEqual(len(model4.aliasing), 0)

        model5 = lm("y ~ d + I(np.power(d, 2))", data=expt4, alias_threshold=0.99)
        self.assertEqual(len(model5.aliasing), 2)

    def test_realworld_coded(self):
        """
        Test conversion between real-world and coded units.
        """
        c1 = c(2.5, 3, 2.5, 3, center=2.75, range=[2.5, 3], name="cement")
        c1_coded = c1.to_coded()
        self.assertListEqual(c1_coded.to_list(), [-1.0, 1.0, -1.0, 1.0])

        c2 = c(
            [-1.0, -1.0, +1.0, +1.0],
            center=2.75,
            range=[2.5, 3],
            name="cement",
            coded=True,
        )
        c2_rw = c2.to_realworld()
        self.assertListEqual(c2_rw.to_list(), [2.5, 2.5, 3.0, 3.0])


if __name__ == "__main__":
    unittest.main()



"""
Class for ControlChart: robust control charts with a balance between CUSUM and Shewhart properties.
"""

import numpy as np
import pandas as pd

from ..univariate.metrics import median_abs_deviation
from ..regression.methods import repeated_median_slope


def rho(x, k=2.52):
    """
    Bi-weight rho function.

    Fixed constant of k=2.52 is from p 289 of the paper
    https://onlinelibrary.wiley.com/doi/abs/10.1002/for.1125
    """
    return k if np.abs(x) > k else k * (1 - np.power(1 - np.power(x / k, 2), 3))


def psi(x, k=2.0):
    """
    Pre-cleaning based on the Huber y-function can be interpreted as replacing unexpected
    high or low values by a more likely value.
    From p 288 of the paper
    https://onlinelibrary.wiley.com/doi/abs/10.1002/for.1125
    """
    return x if abs(x) < k else k * np.sign(x)


class ControlChart(object):
    """
    Creates control chart instance objects.
    """

    def __init__(self, style="robust", variant="HW"):
        """
        Create/initialize a control chart.


        Args: style (str, optional): Which style control chart to calculate. Defaults to "robust".
            Other choice is 'regular' (i.e. not-robust) calculations. User should then ensure that
            no outliers are present in the data.

            variant (str, optional): Many variants of control charts are available.

                The default is a Holt-Winters (HW) chart, with automatic determination of control
                chart parameters. This chart is a blend of infinite history (CUSUM) charts, and an
                instantaneous (no history taken into account) Shewhart chart. The exact blend is
                specified by parameters `ld_1` (lambda 1) and `ld_2` (lambda 2).

                Other variants are:

                'xbar.no.subgroup' [Shewhart chart, with no subgroups]. In other words, each
                observation is independently plotted on the control chart.

                'CUSUM' (CUmulative SUM) chart, which uses all the history of the chart.
        """
        self.style = style.strip()
        self.variant = variant.strip().lower()

        # Will be calculated by the self.fit_limits() function
        self.target = None
        self._given_target = None
        self._given_s = None
        self.s = None
        # index of elements which are found to be outside +/- 3S
        self.idx_outside_3S = []
        self.warm_up = {}

        columns = [
            "y",
            "psi_input",
            "rho_input",
            "y_star",
            "alpha_hat",
            "beta_hat",
            "sigma_hat",
            "error",
        ]
        self.df = pd.DataFrame(columns=columns, dtype=np.float64)

    def calculate_limits(self, y, target=None, s=None, **kwargs):
        """
        Finds, for a given vector `y`, the control chart target and limits.

        Only for the Holt-Winters method, and only when there are more than
            min(20, max(10, np.ceil(0.10 * N))))

        measurements, where N is the length of the input vector. In other words, the provided
        target and standard deviation are only used if more than 10 to 20 measurements.
        If `target` and `s` are numeric, then that target value and that standard deviation value
        are used, otherwise these values are estimated.
        """
        self._given_target = target
        self._given_s = s

        if s is not None:
            self.s = float(s)
            assert (
                s > 0.0
            ), "The given standard deviation cannot be zero, and must be positive."
            assert s < 1e300, "The given standard deviation cannot be this large."

        if target is not None:
            self.target = float(target)

        self.df["y"] = y.ravel()
        self.N = self.df.shape[0]

        # Between M = 10 and 20 samples required to warm-up (calculate summary statistics)
        self.warm_up["M"] = int(min(20, max(10, np.ceil(0.10 * self.N))))

        if (self.warm_up["M"] > self.N) and self.variant.strip().lower() == "hw":
            # TO CHECK: Completely handle the case with very few samples. Is everything filled in?
            # Also check case when some of these samples are NAN, you might have even fewer still.
            self.target = self._target_calculated_best = self.df["y"].median()
            self.s = self._tau = self.df["y"].std()
            self.df["y_star"] = self.df["y"].values
            self.df["alpha_hat"] = self.target
            self.df["beta_hat"] = 0
            self.df["sigma_hat"] = np.nan
            self.df["error"] = np.nan
            return

        # Check if there are enough training samples:
        if self.N < 2 * self.warm_up["M"]:
            self.train_samples = list(np.arange(0, self.N))
        else:
            self.train_samples = list(np.arange(self.warm_up["M"], self.N))

        for key, val in kwargs.items():
            setattr(self, key, val)

        if self.variant.strip().lower() == "hw":
            if not hasattr(self, "ld_1"):
                setattr(self, "ld_1", None)

            if not hasattr(self, "ld_2"):
                setattr(self, "ld_2", None)

            self._holt_winters_parameter_fit()

        if self.variant.strip().lower() == "xbar.no.subgroup":
            self._xbar_no_subgroup_fit()

        # After whichever fit is completed, check which are outside +/- 3S:
        idx_bool = (self.df["y"] - self.target).abs() > 3.0 * self.s
        self.idx_outside_3S = np.nonzero(idx_bool.ravel())[0].tolist()

    def _xbar_no_subgroup_fit(self):
        """
        The control chart is fit from the data samples, assuming each sample is its own subgroup.
        Variant = 'regulur' | 'robust' switchs how the average and standard deviation are
        calculated.

        Control chart limits assume the data are normally distributed and independent. In
        particular, this last assumption can have consequences if not actually met. Limits may be
        too wide, or too narrow.

        """
        if self.style == "regular":
            self.target = self.df["y"].mean()
            self.s = self.df["y"].std()
        elif self.style == "robust":
            self.target = self.df["y"].median()
            self.s = (self.df["y"] - self.target).abs().median() * 1.4826

    def _holt_winters_parameter_fit(self):
        """
        Recommended in the paper: not to fit the lambda_s value, but to use a grid search for the
        lambda_1 and lambda_2 values. This is done in a 5x5 grid in the code below.
        """
        self.ld_s = ld_s = 0.2
        rho_func = np.vectorize(rho)

        if self.ld_1 and self.ld_2:
            # User has provided their own lambda_1 and lambda_2 values.
            assert self.ld_1 >= 0.0, "Lambda_1 must be greater than or equal to zero."
            assert self.ld_2 >= 0.0, "Lambda_2 must be greater than or equal to zero."
            assert self.ld_s >= 0.0, "Lambda_s must be greater than or equal to zero."
            assert self.ld_1 <= 1.0, "Lambda_1 must be less than or equal to 1.0."
            assert self.ld_2 <= 1.0, "Lambda_2 must be less than or equal to 1.0."
            assert self.ld_s <= 1.0, "Lambda_s must be less than or equal to 1.0."
            self._holt_winters_warmup_fit(
                ld_1=self.ld_1, ld_2=self.ld_2, ld_s=self.ld_s
            )

        else:
            # User wants to find an value for ld_1 and ld_2 that best fits the data

            ld_1_index = np.linspace(0.1, 0.9, num=5, endpoint=True)
            ld_2_index = np.linspace(0.1, 0.9, num=5, endpoint=True)
            residuals, _ = np.meshgrid(ld_1_index, ld_2_index)

            for i, ld_1 in enumerate(ld_1_index):
                for j, ld_2 in enumerate(ld_2_index):
                    self._holt_winters_warmup_fit(ld_1=ld_1, ld_2=ld_2, ld_s=ld_s)

                    # Apply equation 16 from the paper to the residuals in the 'training' period,
                    # that is the samples after the warm-up period.
                    future_errors = self.df["error"][self.train_samples]
                    S_T_median_error = 1.48 * np.median(abs(future_errors))
                    residuals[i, j] = np.power(S_T_median_error, 2) * np.average(
                        rho_func(future_errors / S_T_median_error)
                    )
                    # print(f"{i}, {j}, {residuals[i, j]:.5f}")

            min_idx = np.argmin(residuals)
            best_ld_1 = ld_1_index[np.unravel_index(min_idx, residuals.shape)[0]]
            best_ld_2 = ld_2_index[np.unravel_index(min_idx, residuals.shape)[1]]
            # Store the parameters that were calculated, even if a `target` or `s` were provided.
            self.ld_1 = best_ld_1
            self.ld_2 = best_ld_2
            self._residuals_HW = residuals
            self._holt_winters_warmup_fit(ld_1=best_ld_1, ld_2=best_ld_2, ld_s=ld_s)

        # Common code for both branches of if-else above
        future_errors = self.df["error"][self.train_samples]
        S_T_median_error = 1.48 * future_errors.abs().median()  # must handle NaNs!
        resids = np.power(S_T_median_error, 2) * np.nanmean(
            rho_func(future_errors / S_T_median_error)
        )

        # Ensure no negative square root is taken
        self._tau = np.sqrt(max(0.0, resids))
        if self.target is None:
            # Estimate the target as the median of the y-star (cleaned) y-values
            self.target = self.df["y_star"].median()
        else:
            self._target_calculated_best = self.df["y_star"].median()

        if self.s is None:
            self.s = (
                self._tau
            )  # or an alternative: self.df["sigma_hat"] is approximately OK

        # The "delta" emphasizes that it is the deviation from the target.
        self._delta_UCL_3sigma = +3.0 * self._tau
        self._delta_LCL_3sigma = -3.0 * self._tau

    def _holt_winters_warmup_fit(self, ld_1=0.5, ld_2=0.8, ld_s=0.2):
        """
        See paper: https://onlinelibrary.wiley.com/doi/abs/10.1002/for.1125

        Calculates the Holt-Winters fitting and control chart parameters, for given values of the
        smoothing parameters lambda_1 (how must local history for the level is used, with values
        approaching 1.0 implying that less history is used), and lambda_2 (history for the trend
        that is used, with lambda_2 approaching 1.0 implying that historical data is less
        interesting), and lambda_s, a similar parameter for the moving variance of the sequence.

        lambda_1 = ld_1 = 0.5 (default): value must be between 0 <= ld_1 <= 1.0
        lambda_2 = ld_2 = 0.8 (default): value must be between 0 <= ld_2 <= 1.0
        lambda_s = ld_s = 0.2 (default): value must be between 0 <= ld_2 <= 1.0, based on values
                                         used in the paper, recommended on page 291.

        The ideal lambda values (ld_1, ld_2, ld_s) can be found from a grid search.
        """
        df = self.df
        y_warm_up = df["y"].iloc[0 : self.warm_up["M"]]
        self.warm_up["y_zero_robust"] = y_warm_up.median()

        if isinstance(self.target, float):
            self.warm_up["alpha_0"] = self.target
            self.warm_up["beta_0"] = 0.0
        else:
            # p 290 of the paper, https://onlinelibrary.wiley.com/doi/abs/10.1002/for.1125
            self.warm_up["beta_0"] = repeated_median_slope(
                np.arange(self.warm_up["M"]), y_warm_up
            )
            self.warm_up["alpha_0"] = np.nanmedian(
                y_warm_up - self.warm_up["beta_0"] * np.arange(self.warm_up["M"])
            )

        if isinstance(self.s, float):
            self.warm_up["sigma_0"] = self.s

        else:
            # p 290 of the paper, https://onlinelibrary.wiley.com/doi/abs/10.1002/for.1125
            warm_up_residuals = (
                y_warm_up - self.warm_up["alpha_0"] - self.warm_up["beta_0"]
            )

            # Some other method that does not rely on SciPy for 1 function.
            self.warm_up["sigma_0"] = median_abs_deviation(
                warm_up_residuals, nan_policy="omit"
            )
            self.warm_up["residuals"] = warm_up_residuals

        df.loc[0, "rho_input"] = (
            self.warm_up["y_zero_robust"]
            - self.warm_up["alpha_0"]
            - self.warm_up["beta_0"]
        ) / self.warm_up["sigma_0"]
        df.loc[0, "psi_input"] = df["rho_input"][0]
        df.loc[0, "y_star"] = self.warm_up["y_zero_robust"]
        df.loc[0, "alpha_hat"] = self.warm_up["alpha_0"]
        df.loc[0, "beta_hat"] = self.warm_up["beta_0"]
        df.loc[0, "sigma_hat"] = self.warm_up["sigma_0"]

        for i in range(1, self.N):
            # Cover the warm-up period, and the rest of the data set. We need that for the residual
            # calculation later anyway.

            # Error = observed - predicted. Predicted = one-step-ahead prediction
            error_i = df["y"][i] - (df["alpha_hat"][i - 1] + df["beta_hat"][i - 1])
            if np.isnan(error_i):
                # If there is an error, replace it with the median of the last 10 error estimates
                # or as many points as available.
                error_i = df["error"].iloc[max(i - 10, 0) : i].abs().median()
            rho_i = error_i / df["sigma_hat"][i - 1]
            prior_variance = np.power(df["sigma_hat"][i - 1], 2)
            sigma_i = np.sqrt(
                rho(rho_i) * ld_s * prior_variance + (1.0 - ld_s) * prior_variance
            )
            psi_i = error_i / sigma_i
            y_star_i = (
                psi(psi_i) * sigma_i + df["alpha_hat"][i - 1] + df["beta_hat"][i - 1]
            )
            alpha_i = ld_1 * y_star_i + (1 - ld_1) * (
                df["alpha_hat"][i - 1] + df["beta_hat"][i - 1]
            )
            beta_i = (
                ld_2 * (alpha_i - df["alpha_hat"][i - 1])
                + (1 - ld_2) * df["beta_hat"][i - 1]
            )
            df.loc[i] = [
                df["y"][i],
                psi_i,
                rho_i,
                y_star_i,
                alpha_i,
                beta_i,
                sigma_i,
                error_i,
            ]

            # Checks: for algorithm debugging:
            # future_errors = df["error"]
            # S_T_median_error = 1.48 * future_errors.abs().median()  # must handle NaNs!
            # resids = np.power(S_T_median_error, 2) * \
            #                      np.nanmean((future_errors / S_T_median_error).apply(rho))
            # print(np.sqrt(max(0.0, resids)))



import numpy as np
import pandas as pd
from ..univariate.metrics import Sn


def calculate_Cpk(
    df: pd.DataFrame,
    which_column: str,
    specifications=(np.NaN, np.NaN),
    trim_percentile: float = 5.0,
) -> float:
    """
    Calculate the process capability, Cpk, near either the lower or the upper limit [will be
    automatically determined which].

    Process capability, nearer the lower limit = (avg - lower_spec)/(3 x std deviation)
    Process capability, nearer the upper limit = (upper_spec - avg)/(3 x std deviation)

    Parameters
    ----------
    df : pd.DataFrame
        Raw data, at least one column is numeric.
    which_column : str
        Indicates which is the column of data that should be used for the Cpk calculation.
    specifications : tuple, optional
        Either a value, if the specification is constant over time; if the specification changes
        over time, then use two column names here, one of which is the lower specification and
        the second is the upper specification.
    trim_percentile : float, optional
        If non-zero, then robust alternatives are used. The value specified here is the percentile
        of the data that is trimmed away; by default 5 percent on the left, and 5% on the right.

    Returns
    -------
    float
        The Cpk value.
    """
    assert trim_percentile >= 0
    assert trim_percentile < 40  # typically a max of 10 to 20 is advised.
    lower_spec, upper_spec = specifications

    if lower_spec is None:
        Cpk_lower_spec = float(
            np.nanpercentile(df[which_column].values, [trim_percentile])
        )
    elif isinstance(lower_spec, str):
        Cpk_lower_spec = df[lower_spec]
    else:
        Cpk_lower_spec = float(lower_spec)

    if upper_spec is None:
        Cpk_upper_spec = float(
            np.nanpercentile(df[which_column].values, [100 - trim_percentile])
        )
    elif isinstance(upper_spec, str):
        Cpk_upper_spec = df[upper_spec]
    else:
        Cpk_upper_spec = float(upper_spec)

    metric_lower = df[which_column] - Cpk_lower_spec
    metric_upper = Cpk_upper_spec - df[which_column]

    if trim_percentile > 0:
        center_lower, center_upper = metric_lower.median(), metric_upper.median()
        spread_lower, spread_upper = Sn(metric_lower), Sn(metric_upper)
    else:
        center_lower, center_upper = metric_lower.mean(), metric_upper.mean()
        spread_lower, spread_upper = metric_lower.std(), metric_upper.std()

    # TODO: return the RSD also: rsd = (spread / center) * 100
    Cpk = np.nanmin(
        [center_lower / (3 * spread_lower), center_upper / (3 * spread_upper)]
    )
    return Cpk



"""
Getting data into the required format for use with this library.

There are 3 useful ways to represent batch data.

`dict`: as a Python dictionary

    data = {
        "batch 1": data frame with varying number of rows, but same number of columns.
        "batch 2": etc
    }

    The `keys` are unique identifiers for each batch, such as integers, or strings.

`melt`: as a single Pandas data frame

    data = pd.DataFrame(...)

        * very large number of rows, for all batches stacked vertically on top of each other
        * some number of columns, one column per tag
        * one column, usually called "batch_id", indicates what the batch number is for that row
        * another column, usually called "time", indicates what the time is within that batch
        * typically sorted, but does not have to be

`wide`: as a single Pandas data frame, as for the "melted" version, but pivoted instead, and
these `wide` dataframes *always* have a multilevel column index, to help distinguish the tags
from the time. It is a requirement of course that this representation is only for aligned data.

    data = pd.DataFrame(...)

        * each row is a unique batch number
        * the multilevel column index has
            * level 0: the column name
            * level 1: the aligned time

        This form only makes sense if the data are aligned, so that there is the same number of
        unique elements in the level-1 column index.
"""

import numpy as np
import pandas as pd


def check_valid_batch_dict(in_dict: dict, no_nan=False) -> bool:
    """Check if the incoming dictionary of batch data is a valid dictionary of data.

    Checks:
    1. All batches in the dictionary have the same number of columns.
    2. All columns are numeric.
    3. If `no_nan` is True, also checks that there are no NaNs.

    Parameters
    ----------
    in_dict : dict
        A dictionary of batch data.

    no_nan : bool
        If True, will also check that no missing values are present.


    Returns
    -------
    bool
        True, if it passes the checks.
    """
    assert len(in_dict) > 1, "At least 1 batch is required in the dataframe dictionary."
    batch1 = in_dict[list(in_dict.keys())[0]]
    base_columns = set(batch1.columns)
    check = True
    for bid, batch in in_dict.items():
        # Check 1
        check = check & (base_columns == set(batch.columns))
        assert (
            check
        ), f"The column names must be the same in all batches. Differs in {bid}."

        # Check 2
        check *= batch.select_dtypes(include=[np.number]).shape[1] == batch.shape[1]
        assert check, f"All columns must be a numeric type. Differs in {bid}."

        # Check 3
        if no_nan:
            check *= batch.isna().values.sum() == 0
            assert check, f"No missing values allowed. Missing values found in {bid}."

    return bool(check)


def dict_to_melted(
    in_df: pd.DataFrame, insert_batch_id_column=True, insert_sequence_column=False
) -> pd.DataFrame:
    """Reverse of `melted_to_dict`"""
    out_df = pd.DataFrame()
    batch_id_col = "batch_id"

    num_rows = 0
    for idx, (batch_id, batch) in enumerate(in_df.items()):
        if idx == 0:
            num_rows = batch.shape[0]
            sequence = np.arange(0, num_rows)
        assert (
            num_rows == batch.shape[0]
        ), "All batches must have the same number of samples"

        if insert_batch_id_column and batch_id_col not in batch:
            batch.insert(0, batch_id_col, batch_id)

        if insert_sequence_column:
            batch.insert(0, "_sequence_", sequence)

        out_df = out_df.append(batch)

    return out_df


def dict_to_wide(in_df: dict, group_by_batch=False) -> pd.DataFrame:
    """
    Data must be aligned (warped) already so that every batch has the same number of *rows*!

    `group_by_batch`, if True, means that all the data from the first batch is on the left
    of the output dataframe, and the last batch is collected on the right.

    If `group_by_batch` is False, then data for the same tag are grouped together, side-by-side.
    """
    out_df = dict_to_melted(
        in_df=in_df, insert_batch_id_column=True, insert_sequence_column=True
    )
    aligned_wide_df = out_df.pivot(index="batch_id", columns="_sequence_")
    if group_by_batch:
        pass
        # TODO: use the hierarchical indexing and regroup the columns

    return aligned_wide_df


def melted_to_dict(in_df: pd.DataFrame, batch_id_col) -> dict:
    """
    Loads a "melted" data set, where one of the columns is the `batch_id_col`.
    The data are grouped along the unique values of `batch_id_col`, and each group is stored
    in a dictionary. The dictionary keys are the batch identifier, and the corresponding value
    is a Pandas dataframe of the batch data for that batch.
    """
    batches = {}
    assert (
        batch_id_col in in_df
    ), "The `batch_id_col` column does not exist in the incoming dataframe."
    for batch_id, batch in in_df.groupby(batch_id_col):
        batches[batch_id] = batch

    return batches


def melted_to_wide(in_df: pd.DataFrame, batch_id_col) -> dict:
    """
    Data must be aligned already.
    """
    assert batch_id_col in in_df
    return {}

    # max_places = int(np.ceil(np.log10(aligned_df["_sequence_"].max())))
    # aligned_wide_df = aligned_df.pivot(index="batch_id", columns="_sequence_")
    # new_labels = [
    #     "-".join(item)
    #     for item in zip(
    #         aligned_wide_df.columns.get_level_values(0),
    #         [str(val).zfill(max_places) for val in aligned_wide_df.columns.get_level_values(1)],
    #     )
    # ]
    # aligned_wide_df.columns = new_labels
    # TODO: add the column multilevel column index.
    # return dict_to_wide(melted_to_dict(in_df, batch_id_col))


def wide_to_melted(in_df: pd.DataFrame) -> pd.DataFrame:
    # dict_to_melted(dict_to_wide(in_df))
    return pd.DataFrame()


def wide_to_dict():
    pass


def melt_df_to_series(
    in_df: pd.DataFrame, exclude_columns=["batch_id"], name=None
) -> pd.Series:
    """Returns a Series with a multilevel-index, melted from the DataFrame"""
    out = in_df.drop(exclude_columns, axis=1).T.stack()
    out.name = name
    return out
